"""Base module of AS-AP-OpticsCorr IOC."""

import time as _time
from copy import deepcopy as _dcopy
from threading import Thread as _Thread
import numpy as _np
from epics import PV as _PV

from .. import util as _util
from ..callbacks import Callback as _Callback
from ..envars import VACA_PREFIX as _vaca_prefix
from ..namesys import SiriusPVName as _PVName
from ..search import LLTimeSearch as _LLTimeSearch
from ..clientconfigdb import ConfigDBClient as _ConfigDBClient, \
    ConfigDBException as _ConfigDBException
from ..pwrsupply.csdev import Const as _PSConst
from ..timesys.csdev import Const as _TIConst, \
    get_hl_trigger_database as _get_trig_db

from .opticscorr import OpticsCorr as _OpticsCorr
from .csdev import Const as _Const, \
    get_chrom_database as _get_chrom_database, \
    get_tune_database as _get_tune_database
from .utils import HandleConfigNameFile as _HandleConfigNameFile

# Constants
ALLSET = 0x1f
ALLCLR_SYNCON = 0x00
ALLCLR_SYNCOFF = 0x10


class BaseApp(_Callback):
    """Main application for handling optics correction."""

    _optics_param = ''

    def __init__(self, acc):
        """Class constructor."""
        super().__init__()
        self._acc = acc.upper()

        # consts
        if self._optics_param == 'tune':
            self._pvs_database = _get_tune_database(self._acc)
            if self._acc == 'BO':
                self._psfams = _Const.BO_QFAMS_TUNECORR
                self._psfam_nelm = _Const.BO_QFAMS_NELM
            else:
                self._psfams = _Const.SI_QFAMS_TUNECORR
                self._psfam_nelm = _Const.SI_QFAMS_NELM
                self._trigger_name = _PVName('SI-Glob:TI-Mags-Quads')
                self._event_name = 'TunSI'
        else:
            self._pvs_database = _get_chrom_database(self._acc)
            if self._acc == 'BO':
                self._psfams = _Const.BO_SFAMS_CHROMCORR
                self._psfam_nelm = _Const.BO_SFAMS_NELM
            else:
                self._psfams = _Const.SI_SFAMS_CHROMCORR
                self._psfam_nelm = _Const.SI_SFAMS_NELM
                self._trigger_name = _PVName('SI-Glob:TI-Mags-Sexts')
                self._event_name = 'ChromSI'

        self._status = ALLSET if self._acc == 'SI' else 0x0f

        self._optprm_est = [0.0, 0.0]

        self._apply_corr_cmd_count = 0
        self._config_ps_cmd_count = 0

        self._psfam_check_connection = {fam: 0 for fam in self._psfams}
        self._psfam_check_pwrstate_sts = {fam: 0 for fam in self._psfams}
        self._psfam_check_opmode_sts = {fam: -1 for fam in self._psfams}
        self._psfam_check_ctrlmode_mon = {fam: 1 for fam in self._psfams}
        self._psfam_intstr_rb = {fam: 0 for fam in self._psfams}

        if self._acc == 'SI':
            self._corr_method = _Const.CorrMeth.Proportional
            self._corr_group = _Const.CorrGroup.TwoKnobs
            self._sync_corr = _Const.SyncCorr.Off

            self._config_ti_cmd_count = 0
            self._timing_check_config = 9*[0]

            self._measuring_config = False
            self._meas_config_status = _Const.MeasMon.Idle
            self._meas_config_lastcomm = _Const.MeasCmd.Reset
            self._meas_config_wait = 1
            self._meas_config_name = 'UNDEF'
            self._meas_config_2_save = None
            self._meas_config_save_cmd_count = 0

            self._is_storedebeam = 0
        else:
            self._corr_method = _Const.CorrMeth.Additional
            self._corr_group = _Const.CorrGroup.TwoKnobs
            self._sync_corr = _Const.SyncCorr.Off

        # Get focusing and defocusing families
        psfam_focusing = []
        psfam_defocusing = []
        ps_focus_type = 'QF' if self._optics_param == 'tune' else 'SF'
        for fam in self._psfams:
            if ps_focus_type in fam:
                psfam_focusing.append(fam)
            else:
                psfam_defocusing.append(fam)
        psfam_focusing = tuple(psfam_focusing)
        psfam_defocusing = tuple(psfam_defocusing)

        # Initialize correction parameters from local file and configdb
        self.cn_handler = _HandleConfigNameFile(self._acc, self._optics_param)
        self.cdb_client = _ConfigDBClient(
            config_type=self._acc.lower()+'_'+self._optics_param+'corr_params')
        [done, corrparams] = self._get_corrparams()
        if done:
            self._config_name = corrparams[0]
            self._nominal_matrix = corrparams[1]
            self._psfam_nom_intstr = corrparams[2]
            self._nominal_opticsparam = corrparams[3]
            self._opticscorr = _OpticsCorr(
                magnetfams_ordering=self._psfams,
                nominal_matrix=self._nominal_matrix,
                nominal_intstrengths=self._psfam_nom_intstr,
                nominal_opticsparam=self._nominal_opticsparam,
                magnetfams_focusing=psfam_focusing,
                magnetfams_defocusing=psfam_defocusing)
        else:
            raise Exception(
                "Could not read correction parameters from configdb.")

        # Connect to Quadrupoles Families
        self._psfam_intstr_sp_pvs = dict()
        self._psfam_intstr_rb_pvs = dict()
        self._psfam_pwrstate_sel_pvs = dict()
        self._psfam_pwrstate_sts_pvs = dict()
        self._psfam_opmode_sel_pvs = dict()
        self._psfam_opmode_sts_pvs = dict()
        self._psfam_ctrlmode_mon_pvs = dict()
        for fam in self._psfams:
            pss = _PVName(self._acc+'-Fam:PS-'+fam).substitute(
                prefix=_vaca_prefix)
            intstr = 'KL' if self._optics_param == 'tune' else 'SL'
            self._psfam_intstr_sp_pvs[fam] = _PV(
                pss.substitute(propty_name=intstr, propty_suffix='SP'),
                connection_callback=self._callback_conn_psfam,
                connection_timeout=0.05)

            self._psfam_pwrstate_sel_pvs[fam] = _PV(
                pss.substitute(propty_name='PwrState', propty_suffix='Sel'),
                connection_timeout=0.05)
            self._psfam_pwrstate_sts_pvs[fam] = _PV(
                pss.substitute(propty_name='PwrState', propty_suffix='Sts'),
                callback=self._callback_psfam_pwrstate_sts,
                connection_timeout=0.05)

            self._psfam_opmode_sel_pvs[fam] = _PV(
                pss.substitute(propty_name='OpMode', propty_suffix='Sel'),
                connection_timeout=0.05)
            self._psfam_opmode_sts_pvs[fam] = _PV(
                pss.substitute(propty_name='OpMode', propty_suffix='Sts'),
                callback=self._callback_psfam_opmode_sts,
                connection_timeout=0.05)

            self._psfam_ctrlmode_mon_pvs[fam] = _PV(
                pss.substitute(propty_name='CtrlMode', propty_suffix='Mon'),
                callback=self._callback_psfam_ctrlmode_mon,
                connection_timeout=0.05)

        if self._acc == 'SI':
            # Connect to Timing
            evg = _LLTimeSearch.get_evg_name()
            trig_name = self._trigger_name.substitute(prefix=_vaca_prefix)
            evt_name = _PVName(evg).substitute(
                prefix=_vaca_prefix, propty_name=self._event_name)
            try:
                trig_db = _get_trig_db(trig_name)
                self._evt_src_idx = trig_db['Src-Sel']['enums'].index(evt_name)
            except (KeyError, ValueError):
                self._evt_src_idx = 1

            self._trigger_state_sel = _PV(
                trig_name.substitute(propty='State-Sel'),
                connection_timeout=0.05)
            self._trigger_state_sts = _PV(
                trig_name.substitute(propty='State-Sts'),
                callback=self._callback_timing_state, connection_timeout=0.05)

            self._trigger_polarity_sel = _PV(
                trig_name.substitute(propty='Polarity-Sel'),
                connection_timeout=0.05)
            self._trigger_polarity_sts = _PV(
                trig_name.substitute(propty='Polarity-Sts'),
                callback=self._callback_timing_state, connection_timeout=0.05)

            self._trigger_src_sel = _PV(
                trig_name.substitute(propty='Src-Sel'),
                connection_timeout=0.05)
            self._trigger_src_sts = _PV(
                trig_name.substitute(propty='Src-Sts'),
                callback=self._callback_timing_state, connection_timeout=0.05)

            self._trigger_nrpulses_sp = _PV(
                trig_name.substitute(propty='NrPulses-SP'),
                connection_timeout=0.05)
            self._trigger_nrpulses_rb = _PV(
                trig_name.substitute(propty='NrPulses-RB'),
                callback=self._callback_timing_state, connection_timeout=0.05)

            self._trigger_duration_sp = _PV(
                trig_name.substitute(propty='Duration-SP'),
                connection_timeout=0.05)
            self._trigger_duration_rb = _PV(
                trig_name.substitute(propty='Duration-RB'),
                callback=self._callback_timing_state, connection_timeout=0.05)

            self._trigger_delay_sp = _PV(
                trig_name.substitute(propty='Delay-SP'),
                connection_timeout=0.05)
            self._trigger_delay_rb = _PV(
                trig_name.substitute(propty='Delay-RB'),
                callback=self._callback_timing_state, connection_timeout=0.05)

            self._event_mode_sel = _PV(
                evt_name.substitute(propty=evt_name.propty+'Mode-Sel'),
                connection_timeout=0.05)
            self._event_mode_sts = _PV(
                evt_name.substitute(propty=evt_name.propty+'Mode-Sts'),
                callback=self._callback_timing_state, connection_timeout=0.05)

            self._event_delaytype_sel = _PV(
                evt_name.substitute(propty=evt_name.propty+'DelayType-Sel'),
                connection_timeout=0.05)
            self._event_delaytype_sts = _PV(
                evt_name.substitute(propty=evt_name.propty+'DelayType-Sts'),
                callback=self._callback_timing_state, connection_timeout=0.05)

            self._event_delay_sp = _PV(
                evt_name.substitute(propty=evt_name.propty+'Delay-SP'),
                connection_timeout=0.05)
            self._event_delay_rb = _PV(
                evt_name.substitute(propty=evt_name.propty+'Delay-RB'),
                callback=self._callback_timing_state, connection_timeout=0.05)

            self._event_exttrig_cmd = _PV(
                evt_name.substitute(propty=evt_name.propty+'ExtTrig-Cmd'),
                connection_timeout=0.05)

            # Connect to CurrInfo
            self._storedebeam_pv = _PV(
                _PVName('SI-Glob:AP-CurrInfo:StoredEBeam-Mon').substitute(
                    prefix=_vaca_prefix),
                callback=self._callback_get_storedebeam,
                connection_timeout=0.05)

            # Connect to Tunes
            self._tune_x_pv = _PV(
                _PVName('SI-Glob:DI-Tune-H:TuneFrac-Mon').substitute(
                    prefix=_vaca_prefix),
                connection_timeout=0.05)
            self._tune_y_pv = _PV(
                _PVName('SI-Glob:DI-Tune-V:TuneFrac-Mon').substitute(
                    prefix=_vaca_prefix),
                connection_timeout=0.05)

        # Create map of pv -> write function
        self.map_pv2write = {
            'ApplyDelta-Cmd': self.cmd_apply_corr,
            'ConfigName-SP': self.set_config_name,
            'CorrMeth-Sel': self.set_corr_meth,
            'CorrGroup-Sel': self.set_corr_group,
            'SyncCorr-Sel': self.set_sync_corr,
            'ConfigPS-Cmd': self.cmd_config_ps,
            'ConfigTiming-Cmd': self.cmd_config_ti,
            'MeasConfigWait-SP': self.set_meas_config_wait,
            'MeasConfig-Cmd': self.cmd_meas_config,
            'MeasConfigName-SP': self.set_meas_config_name,
            'MeasConfigSave-Cmd': self.cmd_meas_config_save,
        }

    def init_database(self):
        """Set initial PV values."""
        self.run_callbacks('ConfigName-SP', self._config_name)
        self.run_callbacks('ConfigName-RB', self._config_name)
        self.update_corrparams_pvs()
        self.run_callbacks('Log-Mon', 'Started.')
        self.run_callbacks('Status-Mon', self._status)
        if self._optics_param == 'tune':
            for fam in self._psfams:
                self.run_callbacks(
                    'RefKL'+fam+'-Mon', self._psfam_refkl[fam])
            self.run_callbacks('DeltaTuneX-Mon', self._optprm_est[0])
            self.run_callbacks('DeltaTuneY-Mon', self._optprm_est[1])
        else:
            self.run_callbacks('ChromX-Mon', self._optprm_est[0])
            self.run_callbacks('ChromY-Mon', self._optprm_est[1])

    def update_corrparams_pvs(self):
        """Set initial correction parameters PVs values."""
        raise NotImplementedError

    @property
    def pvs_database(self):
        """Return PVs database."""
        return self._pvs_database

    def process(self, interval):
        """Sleep."""
        _time.sleep(interval)

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        if reason in self.map_pv2write.keys():
            status = self.map_pv2write[reason](value)
        else:
            status = False
        return status

    # ------ handle pv write methods -------

    def cmd_apply_corr(self, value):
        """ApplyCorr command."""
        if self._apply_corr():
            self._apply_corr_cmd_count += 1
            self.run_callbacks('ApplyDelta-Cmd', self._apply_corr_cmd_count)
        return False

    def set_config_name(self, value):
        """Set configuration name."""
        [done, corrparams] = self._get_corrparams(value)
        if done:
            try:
                self._opticscorr.nominal_matrix = corrparams[1]
                self._opticscorr.nominal_intstrengths = corrparams[2]
                self._opticscorr.nominal_opticsparam = corrparams[3]
                self._calc_intstrength()
            except Exception:
                self.run_callbacks(
                    'Log-Mon', 'Could not update correction parameters.')
                return False
            else:
                self._config_name = corrparams[0]
                self.cn_handler.set_config_name(self._config_name)
                self.run_callbacks('ConfigName-RB', self._config_name)

                self._nominal_matrix = corrparams[1]
                self._psfam_nom_intstr = corrparams[2]
                self._nominal_opticsparam = corrparams[3]
                self.update_corrparams_pvs()
                self.run_callbacks('Log-Mon', 'Updated correction parameters.')
                return True
        self.run_callbacks('Log-Mon', 'ERR: Config not found in configdb.')
        return False

    def set_corr_meth(self, value):
        """Set CorrMeth."""
        if value == self._corr_method:
            return False
        self._corr_method = value
        self.run_callbacks('CorrMeth-Sts', value)
        self._calc_intstrength()
        return True

    def set_corr_group(self, value):
        """Set CorrGroup."""
        if value == self._corr_group:
            return False
        self._corr_group = value
        self.run_callbacks('CorrGroup-Sts', self._corr_group)
        self._calc_intstrength()
        return True

    def set_sync_corr(self, value):
        """Set SyncCorr."""
        if self._meas_config_status == _Const.MeasMon.Measuring:
            self.run_callbacks(
                'Log-Mon', 'ERR: Configuration measurement in progress.')
            return False
        if value == self._sync_corr:
            return False

        self._sync_corr = value

        if self._config_ps():
            self._config_ps_cmd_count += 1
            self.run_callbacks('ConfigPS-Cmd', self._config_ps_cmd_count)

        if value == 1:
            if self._config_timing():
                self._config_ti_cmd_count += 1
                self.run_callbacks(
                    'ConfigTiming-Cmd', self._config_ti_cmd_count)

        val = 1
        if (self._status & 0x1) == 0:
            for fam in self._psfams:
                self._psfam_check_opmode_sts[fam] = \
                    self._psfam_opmode_sts_pvs[fam].value

            opmode = _PSConst.OpMode.SlowRefSync if value \
                else _PSConst.OpMode.SlowRef
            val = any(op != opmode for op in
                      self._psfam_check_opmode_sts.values())

        self._status = _util.update_bit(
            v=self._status, bit_pos=2, bit_val=val)

        self.run_callbacks('Status-Mon', self._status)
        self.run_callbacks('SyncCorr-Sts', self._sync_corr)
        return True

    def cmd_config_ps(self, value):
        """ConfigPS command."""
        if self._config_ps():
            self._config_ps_cmd_count += 1
            self.run_callbacks('ConfigPS-Cmd', self._config_ps_cmd_count)
        return False

    def cmd_config_ti(self, value):
        """ConfigTiming command."""
        if self._config_timing():
            self._config_ti_cmd_count += 1
            self.run_callbacks('ConfigTiming-Cmd', self._config_ti_cmd_count)
        return False

    def set_meas_config_wait(self, value):
        """Set MeasConfigWait."""
        if value == self._meas_config_wait:
            return False
        self._meas_config_wait = value
        self.run_callbacks('MeasConfigWait-RB', value)
        return True

    def cmd_meas_config(self, value):
        """MeasConfig command."""
        if value == _Const.MeasCmd.Start:
            status = self._start_meas_config()
        elif value == _Const.MeasCmd.Stop:
            status = self._stop_meas_config()
        elif value == _Const.MeasCmd.Reset:
            status = self._reset_meas_config()
        if status:
            self._meas_config_lastcomm = value
        return status

    def set_meas_config_name(self, value):
        """Set MeasConfigName."""
        if value == self._meas_config_name:
            return False
        if not self.cdb_client.check_valid_configname(value):
            self.run_callbacks('Log-Mon', 'ERR: Config name not valid!')
            return False
        self._meas_config_name = value
        self.run_callbacks('MeasConfigName-RB', value)
        return True

    def cmd_meas_config_save(self, value):
        """MeasConfigSave command."""
        if self._meas_config_2_save is None:
            self.run_callbacks(
                'Log-Mon', 'ERR: No new data to save in configdb!')
        elif self._save_corrparams(self._meas_config_name):
            self._meas_config_save_cmd_count += 1
            self.run_callbacks(
                'MeasConfigSave-Cmd', self._meas_config_save_cmd_count)

            self._config_name = _dcopy(self._meas_config_name)
            self.cn_handler.set_config_name(self._config_name)
            self.run_callbacks('ConfigName-RB', self._config_name)

            self._meas_config_name = 'UNDEF'
            self.run_callbacks('MeasConfigName-SP', self._meas_config_name)
            self.run_callbacks('MeasConfigName-RB', self._meas_config_name)

            self.run_callbacks('Log-Mon', 'Updated config. name.')
        return False

    # ---------- auxiliar methods ----------

    def _get_corrparams(self, config_name=''):
        """Get response matrix from configurations database."""
        try:
            if not config_name:
                config_name = self.cn_handler.get_config_name()
            params = self.cdb_client.get_config_value(name=config_name)
        except _ConfigDBException:
            return [False, []]
        else:
            data = list()
            data.append(config_name)
            data.extend(self._handle_corrparams_2_read(params))
            return [True, data]

    def _handle_corrparams_2_read(self, params):
        """Edit correction params."""
        raise NotImplementedError

    def _save_corrparams(self, config_name):
        """Save correction parameters in configdb."""
        value = self._handle_corrparams_2_save()
        try:
            self.cdb_client.insert_config(name=config_name, value=value)
        except _ConfigDBException as err:
            self._meas_config_2_save = value
            if err.server_code == -2:
                log_msg = 'ERR: Could not connect to configdb!'
            else:
                log_msg = 'ERR: Could not save configuration in configdb!'
        else:
            self._meas_config_2_save = None
            log_msg = "Saved config. '{}' in configdb!".format(config_name)
        self.run_callbacks('Log-Mon', log_msg)
        return 'ERR' not in log_msg

    def _handle_corrparams_2_save(self):
        """Handle configuration value to save."""
        raise NotImplementedError

    def _is_status_ok(self):
        """Return if status is OK."""
        if self._sync_corr == _Const.SyncCorr.Off:
            is_ok = self._status in (ALLCLR_SYNCOFF, ALLCLR_SYNCON)
        else:
            is_ok = self._status == ALLCLR_SYNCON
        return is_ok

    def _calc_intstrength(self):
        """Calculate integrated strength."""
        raise NotImplementedError

    def _apply_intstrength(self, int_strengths):
        """Apply integrated strength."""
        for fam, pvobj in self._psfam_intstr_sp_pvs.items():
            pvobj.put(int_strengths[fam])

    def _apply_corr(self):
        """Apply correction."""
        raise NotImplementedError

    def _config_ps(self):
        """Configurate power supplies."""
        opmode = _PSConst.OpMode.SlowRefSync if self._sync_corr \
            else _PSConst.OpMode.SlowRef
        for fam in self._psfams:
            if self._psfam_pwrstate_sel_pvs[fam].connected:
                self._psfam_pwrstate_sel_pvs[fam].put(_PSConst.PwrStateSel.On)
                self._psfam_opmode_sel_pvs[fam].put(opmode)
            else:
                self.run_callbacks('Log-Mon', 'ERR:'+fam+' is disconnected.')
                return False
        self.run_callbacks('Log-Mon', 'Configuration sent to power supplies.')
        return True

    def _config_timing(self):
        """Configurate Timing."""
        conn = not any(pv.connected is False for pv in [
            self._trigger_state_sel,
            self._trigger_polarity_sel,
            self._trigger_src_sel,
            self._trigger_nrpulses_sp,
            self._trigger_duration_sp,
            self._trigger_delay_sp,
            self._event_mode_sel,
            self._event_delaytype_sel,
            self._event_delay_sp])
        if conn:
            self._trigger_state_sel.put(_TIConst.DsblEnbl.Enbl)
            self._trigger_polarity_sel.put(_TIConst.TrigPol.Normal)
            self._trigger_src_sel.put(self._evt_src_idx)
            self._trigger_nrpulses_sp.put(1)
            self._trigger_duration_sp.put(150)
            self._trigger_delay_sp.put(0)
            self._event_mode_sel.put(_TIConst.EvtModes.External)
            self._event_delaytype_sel.put(_TIConst.EvtDlyTyp.Incr)
            self._event_delay_sp.put(0)

            self.run_callbacks('Log-Mon', 'Configuration sent to TI.')
            return True
        else:
            self.run_callbacks('Log-Mon', 'ERR:Some TI PV is disconnected.')
            return False

    def _get_tunes(self):
        """Return tunes."""
        if self._tune_x_pv.connected:
            sts = True
            tunex = self._tune_x_pv.value
            tuney = self._tune_y_pv.value
        else:
            sts = False
            tunex, tuney = [0, 0]
        return sts, [tunex, tuney]

    def _get_optics_param(self):
        """Get nominal optics parameter."""
        raise NotImplementedError

    def _get_delta_intstrength(self, fam):
        """Get delta integrated strengths."""
        raise NotImplementedError

    def _start_meas_config(self):
        """Start configuration measurement."""
        cont = True
        if self._sync_corr == _Const.SyncCorr.On:
            log_msg = 'ERR: Turn off syncronized correction!'
            cont = False
        elif self._meas_config_name == 'UNDEF':
            log_msg = 'ERR: Define a conf.name to save the measure!'
            cont = False
        elif self._status != 0:
            log_msg = 'ERR: Verify power supplies status!'
            cont = False
        elif self._measuring_config:
            log_msg = 'ERR: Measurement already in progress!'
            cont = False
        elif not self._is_storedebeam:
            log_msg = 'ERR: Cannot measure, there is not stored beam!'
            cont = False
        elif not self._tune_x_pv.connected or not self._tune_y_pv.connected:
            log_msg = 'ERR: Cannot measure, tune PVs not connected!'
            cont = False
        if not cont:
            self.run_callbacks('Log-Mon', log_msg)
            return False

        self.run_callbacks(
            'Log-Mon', 'Starting correction config measurement!')
        self._measuring_config = True
        thread = _Thread(target=self._meas_config_thread, daemon=True)
        thread.start()
        return True

    def _stop_meas_config(self):
        """Stop configuration measurement."""
        if not self._measuring_config:
            self.run_callbacks('Log-Mon', 'ERR: No measurement occuring!')
            return False
        self.run_callbacks('Log-Mon', 'Aborting config measurement!')
        self._measuring_config = False
        return True

    def _reset_meas_config(self):
        """Reset configuration measurement."""
        if self._measuring_config:
            self.run_callbacks(
                'Log-Mon', 'WARN: Status not reset, measurement in progress!')
            return False
        self.run_callbacks('Log-Mon', 'Reseting measurement status!')
        self._meas_config_status = _Const.MeasMon.Idle
        self.run_callbacks('MeasConfigStatus-Mon', self._meas_config_status)
        return True

    def _meas_config_thread(self):
        """Configuration measurement thread."""
        self._meas_config_status = _Const.MeasMon.Measuring
        self.run_callbacks('MeasConfigStatus-Mon', self._meas_config_status)

        respm = _np.zeros((2, len(self._psfams)), dtype=float)
        fams_intstr0 = {
            fam: self._psfam_intstr_rb[fam] for fam in self._psfams}
        fams_intstr = _dcopy(fams_intstr0)

        aborted = False
        if self._optics_param == 'chrom':
            sts, data = self._get_optics_param()
            if sts:
                self._nominal_opticsparam = data
            else:
                log_msg = 'ERR: Could not measure chrom!'
                aborted = True

        for fam in self._psfams:
            if not self._measuring_config:
                log_msg = 'Stoped measurement!'
                aborted = True
            elif not self._is_storedebeam:
                log_msg = 'ERR: Stoping measurement, there is no stored beam!'
                aborted = True
            elif self._status != 0:
                log_msg = 'ERR: Stoping measurement, verify power supplies!'
                aborted = True
            elif not self._tune_x_pv.connected or \
                    not self._tune_y_pv.connected:
                log_msg = 'ERR: Stoping measurement, tune PVs not connected!'
                aborted = True

            if aborted:
                break

            fam_idx = self._psfams.index(fam)
            self.run_callbacks(
                'Log-Mon', 'Step: {0:d}/{1:d} --> {2:s}'.format(
                    fam_idx+1, len(self._psfams), fam))

            delta = self._get_delta_intstrength(fam)

            fams_intstr[fam] = fams_intstr0[fam] + delta/2
            self._apply_intstrength(fams_intstr)
            _time.sleep(self._meas_config_wait)
            sts, data = self._get_optics_param()
            if sts:
                param_pos_x, param_pos_y = data
            else:
                log_msg = 'ERR: Could not measure '+self._optics_param+'!'
                aborted = True
                break

            fams_intstr[fam] = fams_intstr0[fam] - delta/2
            self._apply_intstrength(fams_intstr)
            _time.sleep(self._meas_config_wait)
            sts, data = self._get_optics_param()
            if sts:
                param_neg_x, param_neg_y = data
            else:
                log_msg = 'ERR: Could not measure '+self._optics_param+'!'
                aborted = True
                break

            respm[0, fam_idx] = (param_pos_x-param_neg_x)/delta
            respm[1, fam_idx] = (param_pos_y-param_neg_y)/delta

            fams_intstr[fam] = fams_intstr0[fam]
            self._apply_intstrength(fams_intstr)

        self.run_callbacks(
            'Log-Mon', 'Ensure power supplies return to initial values...')
        self._apply_intstrength(fams_intstr0)

        if aborted:
            self._meas_config_status = _Const.MeasMon.Aborted
        else:
            self._meas_config_status = _Const.MeasMon.Completed
            log_msg = 'Correction configuration measurement completed!'

        self.run_callbacks('MeasConfigStatus-Mon', self._meas_config_status)
        self._measuring_config = False
        self.run_callbacks('Log-Mon', log_msg)

        # update corrparams
        self._nominal_matrix = respm.flatten().tolist()
        self._psfam_nom_intstr = [fams_intstr0[fam] for fam in self._psfams]
        self.update_corrparams_pvs()

        try:
            self._opticscorr.nominal_matrix = self._nominal_matrix
            self._opticscorr.nominal_intstrengths = self._psfam_nom_intstr
            self._opticscorr.nominal_opticsparam = self._nominal_opticsparam
            self._calc_intstrength()
            self.run_callbacks('Log-Mon', 'New correction parameters in use.')
        except Exception:
            self._meas_config_2_save = self._handle_corrparams_2_save()
            self.run_callbacks('Log-Mon', 'ERR: Could not use new parameters.')
            self.run_callbacks('Log-Mon', 'ERR: Will not save new parameters.')
        else:
            if self._save_corrparams(self._meas_config_name):
                # update configname
                self._config_name = _dcopy(self._meas_config_name)
                self.cn_handler.set_config_name(self._config_name)
                self.run_callbacks('ConfigName-RB', self._config_name)

                self._meas_config_name = 'UNDEF'
                self.run_callbacks('MeasConfigName-SP', self._meas_config_name)
                self.run_callbacks('MeasConfigName-RB', self._meas_config_name)

    # ---------- callbacks ----------

    def _callback_conn_psfam(self, pvname, conn, **kws):
        """Connection callback."""
        if not conn:
            self.run_callbacks('Log-Mon', 'WARN:'+pvname+' disconnected.')

        fam = _PVName(pvname).dev
        self._psfam_check_connection[fam] = (1 if conn else 0)

        # Change the first bit of correction status
        self._status = _util.update_bit(
            v=self._status, bit_pos=0,
            bit_val=any(v == 0 for v in self._psfam_check_connection.values()))
        self.run_callbacks('Status-Mon', self._status)

    def _callback_psfam_pwrstate_sts(self, pvname, value, **kws):
        """Callback."""
        if value != _PSConst.PwrStateSts.On:
            self.run_callbacks('Log-Mon', 'WARN:'+pvname+' is not On.')

        fam = _PVName(pvname).dev
        self._psfam_check_pwrstate_sts[fam] = value

        # Change the second bit of correction status
        self._status = _util.update_bit(
            v=self._status, bit_pos=1,
            bit_val=any(v != _PSConst.PwrStateSts.On
                        for v in self._psfam_check_pwrstate_sts.values()))
        self.run_callbacks('Status-Mon', self._status)

    def _callback_psfam_opmode_sts(self, pvname, value, **kws):
        """Callback."""
        self.run_callbacks('Log-Mon', 'WARN:'+pvname+' changed.')

        fam = _PVName(pvname).dev
        self._psfam_check_opmode_sts[fam] = value

        # Change the third bit of correction status
        opmode = _PSConst.States.SlowRefSync if self._sync_corr \
            else _PSConst.States.SlowRef
        self._status = _util.update_bit(
            v=self._status, bit_pos=2,
            bit_val=any(v != opmode for v in
                        self._psfam_check_opmode_sts.values()))
        self.run_callbacks('Status-Mon', self._status)

    def _callback_psfam_ctrlmode_mon(self, pvname, value, **kws):
        """Callback."""
        if value != _PSConst.Interface.Remote:
            self.run_callbacks('Log-Mon', 'WARN:'+pvname+' is not Remote.')

        fam = _PVName(pvname).dev
        self._psfam_check_ctrlmode_mon[fam] = value

        # Change the fourth bit of correction status
        self._status = _util.update_bit(
            v=self._status, bit_pos=3,
            bit_val=any(v != _PSConst.Interface.Remote
                        for v in self._psfam_check_ctrlmode_mon.values()))
        self.run_callbacks('Status-Mon', self._status)

    def _callback_timing_state(self, pvname, value, **kws):
        """Callback."""
        if self._trigger_name+':State' in pvname:
            self._timing_check_config[0] = (value == _TIConst.DsblEnbl.Enbl)
        elif self._trigger_name+':Polarity' in pvname:
            self._timing_check_config[1] = (value == _TIConst.TrigPol.Normal)
        elif self._trigger_name+':Src' in pvname:
            self._timing_check_config[2] = (value == self._evt_src_idx)
        elif self._trigger_name+':NrPulses' in pvname:  # 1 pulse
            self._timing_check_config[3] = (value == 1)
        elif self._trigger_name+':Duration' in pvname:  # 150us
            self._timing_check_config[4] = _np.isclose(value, 150, atol=0.1)
        elif self._trigger_name+':Delay' in pvname:  # 0us
            self._timing_check_config[5] = _np.isclose(value, 0, atol=0.1)
        elif self._event_name+'Mode' in pvname:
            self._timing_check_config[6] = value == _TIConst.EvtModes.External
        elif self._event_name+'DelayType' in pvname:
            self._timing_check_config[7] = value == _TIConst.EvtDlyTyp.Incr
        elif self._event_name+'Delay' in pvname:  # 0us
            self._timing_check_config[8] = _np.isclose(value, 0, atol=0.1)

        # Change the fifth bit of correction status
        if self._sync_corr == _Const.SyncCorr.Off:
            bit_val = 0
        else:
            bit_val = any(idx == 0 for idx in self._timing_check_config)
        self._status = _util.update_bit(
            v=self._status, bit_pos=4, bit_val=bit_val)
        self.run_callbacks('Status-Mon', self._status)

    def _callback_get_storedebeam(self, value, **kws):
        """Callback."""
        if value == 0:
            self.run_callbacks('Log-Mon', 'WARN: There is no stored beam!')
        self._is_storedebeam = value



"""PVFake class."""
#  M Newville <newville@cars.uchicago.edu>
#  The University of Chicago, 2010
#  Epics Open License
#
#  ps: code cleanup from LNLS-FAC staff fro pylint adaptation.

import sys
import time
import ctypes
from random import randint as _randint
import copy
from math import log10
import numpy as _np

from epics import dbr
from epics.ca import DEFAULT_CONNECTION_TIMEOUT
from epics.ca import AUTOMONITOR_MAXLENGTH
from epics.ca import HAS_NUMPY

_PVCACHE_ = {}

_PYTYPES = {'float': float, 'int': int, 'bool': int,
            'string': str, 'enum': int, 'char': str}
_CATYPES = {'float': dbr.DOUBLE, 'int': dbr.LONG, 'bool': dbr.INT,
            'string': dbr.STRING, 'enum': dbr.ENUM, 'char': dbr.STRING}

_DATABASE = dict()


def add_to_database(dbase, prefix=''):
    """."""
    if not dbase:
        return
    for key, val in dbase.items():
        _DATABASE.update({prefix+key: copy.deepcopy(val)})


def clear_database():
    """."""
    _DATABASE.clear()


def fmt_time(tstamp=None):
    """Return simple formatter for time values."""
    if tstamp is None:
        tstamp = time.time()
    tstamp, frac = divmod(tstamp, 1)
    return "%s.%5.5i" % (time.strftime("%Y-%m-%d %H:%M:%S",
                                       time.localtime(tstamp)),
                         round(1.e5*frac))


def promote_type(tp, use_time=False, use_ctrl=False):
    """Promote the native field type of a ``chid`` to its TIME or CTRL variant.

    Returns the integer corresponding to the promoted field value.
    """
    ftype = _CATYPES[tp]
    if use_ctrl:
        ftype += dbr.CTRL_STRING
    elif use_time:
        ftype += dbr.TIME_STRING
    if ftype == dbr.CTRL_STRING:
        ftype = dbr.TIME_STRING
    return ftype


def write(msg, newline=True, flush=True):
    """write message to stdout"""
    sys.stdout.write(msg)
    if newline:
        sys.stdout.write("\n")
    if flush:
        sys.stdout.flush()


class PVFake:
    """Fake Epics Process Variable.

    A PV encapsulates an Epics Process Variable.

    The primary interface methods for a pv are to get() and put() is value::

     >>> p = PV(pv_name)  # create a pv object given a pv name
     >>> p.get()          # get pv value
     >>> p.put(val)       # set pv to specified value.

    Additional important attributes in  clude::

     >>> p.pvname         # name of pv
     >>> p.value          # pv value (can be set or g  et)
     >>> p.char_value     # string representation of   pv value
     >>> p.count          # number of elements in array pvs
     >>> p.type           # EPICS data type: 'string','double','enum',
                               # 'long',..
    """

    _fmtsca = ("<PV '%(pvname)s', count=%(count)i, "
               "type=%(typefull)s, access=%(access)s>")
    _fmtarr = ("<PV '%(pvname)s', count=%(count)i/%(nelm)i, "
               "type=%(typefull)s, access=%(access)s>")
    _fields = ('pvname', 'value', 'char_value', 'status', 'ftype', 'chid',
               'host', 'count', 'access', 'write_access', 'read_access',
               'severity', 'timestamp', 'posixseconds', 'nanoseconds',
               'precision', 'units', 'enum_strs',
               'upper_disp_limit', 'lower_disp_limit', 'upper_alarm_limit',
               'lower_alarm_limit', 'lower_warning_limit',
               'upper_warning_limit', 'upper_ctrl_limit', 'lower_ctrl_limit')

    def __init__(self, pvname, callback=None, form='time',
                 verbose=False, auto_monitor=None, count=None,
                 connection_callback=None,
                 connection_timeout=None,
                 access_callback=None):
        """."""
        dbase = _DATABASE.get(pvname)
        if dbase is None:
            raise Exception(
                'PV does not exist in local database. Configure database ' +
                'first with add_to_database module function.')
        self.pvname = pvname.strip()
        self.form = form.lower()
        self.verbose = verbose
        self.auto_monitor = auto_monitor
        self.ftype = None
        self.connected = False
        self.connection_timeout = connection_timeout
        if self.connection_timeout is None:
            self.connection_timeout = DEFAULT_CONNECTION_TIMEOUT
        self._args = {}.fromkeys(self._fields)
        self._args['pvname'] = self.pvname
        self._args['value'] = dbase.get(
            'value', '' if dbase['type'] == 'str' else 0)
        self._args['count'] = count
        self._args['nelm'] = dbase.get('count', 1)
        self._args['type'] = 'unknown'
        self._args['typefull'] = 'unknown'
        self._args['access'] = 'unknown'
        self._args['severity'] = 0  # not simulated correctly
        self._args['status'] = 1  # not simulated correctly
        self._args['precision'] = dbase.get('prec')
        self._args['units'] = dbase.get('unit')
        self._args['upper_disp_limit'] = dbase.get('hilim')
        self._args['upper_disp_limit'] = dbase.get('lolim')
        self._args['upper_alarm_limit'] = dbase.get('high')
        self._args['upper_alarm_limit'] = dbase.get('low')
        self._args['upper_warning_limit'] = dbase.get('hihi')
        self._args['upper_warning_limit'] = dbase.get('lolo')
        self._args['upper_ctrl_limit'] = dbase.get('hilim')
        self._args['upper_ctrl_limit'] = dbase.get('lolim')
        self._args['enum_strs'] = dbase.get('enums')

        self.context = dbr.ECA_NORMAL

        self._pytype = _PYTYPES[dbase['type']]
        self.ftype = promote_type(dbase['type'],
                                  use_ctrl=self.form == 'ctrl',
                                  use_time=self.form == 'time')
        self._args['type'] = dbr.Name(self.ftype).lower()

        self.connection_callbacks = []
        if connection_callback is not None:
            self.connection_callbacks = [connection_callback]

        self.access_callbacks = []
        if access_callback is not None:
            self.access_callbacks = [access_callback]

        self.callbacks = {}
        self._conn_started = False
        if isinstance(callback, (tuple, list)):
            for i, thiscb in enumerate(callback):
                if hasattr(thiscb, '__call__'):
                    self.callbacks[i] = (thiscb, {})
        elif hasattr(callback, '__call__'):
            self.callbacks[0] = (callback, {})

        self._args['chid'] = dbr.chid_t(_randint(1, 10000000))
        self.__on_connect(pvname=pvname, chid=self._args['chid'])
        self.chid = self._args['chid']

        self._set_charval(self._args['value'], call_ca=False)

        self.__on_access_rights_event(read_access=True, write_access=True)

        pvid = (self.pvname, self.form, self.context)
        if pvid not in _PVCACHE_:
            _PVCACHE_[pvid] = self

    def force_connect(self, pvname=None, chid=None, conn=True, **kws):
        """."""
        if chid is None:
            chid = self.chid
        if isinstance(chid, ctypes.c_long):
            chid = chid.value
        self._args['chid'] = self.chid = chid
        self.__on_connect(pvname=pvname, chid=chid, conn=conn, **kws)

    def force_read_access_rights(self):
        """Force a read of access rights.

        not relying on last event callback.
        Note: event callback seems to fail sometimes,
        at least on initial connection on Windows 64-bit.
        """

    def __on_access_rights_event(self, read_access, write_access):
        self._args['read_access'] = read_access
        self._args['write_access'] = write_access

        acc = read_access + 2 * write_access
        access_strs = ('no access', 'read-only', 'write-only', 'read/write')
        self._args['access'] = access_strs[acc]

        for callback in self.access_callbacks:
            if callable(callback):
                callback(read_access, write_access, pv=self)

    def __on_connect(self, pvname=None, chid=None, conn=True):
        """Run callback for connection events."""
        # occassionally chid is still None (ie if a second PV is created
        # while __on_connect is still pending for the first one.)
        # Just return here, and connection will happen later
        if conn:
            self.chid = self._args['chid'] = dbr.chid_t(chid.value)

            # allow reduction of elements, via count argument
            count = self._args['nelm']
            if self._args['count'] is not None:
                count = min(count, self._args['count'])
            self._args['count'] = count

            self._args['host'] = 'localhost:00000'

            _ftype_ = dbr.Name(self.ftype).lower()
            self._args['type'] = _ftype_
            self._args['typefull'] = _ftype_
            self._args['ftype'] = dbr.Name(_ftype_, reverse=True)

            if self.auto_monitor is None:
                self.auto_monitor = count < AUTOMONITOR_MAXLENGTH

        for conn_cb in self.connection_callbacks:
            if hasattr(conn_cb, '__call__'):
                conn_cb(pvname=self.pvname, conn=conn, pv=self)
            elif not conn and self.verbose:
                write("PV '%s' disconnected." % pvname)

        # pv end of connect, force a read of access rights
        self.force_read_access_rights()

        # waiting until the very end until to set self.connected prevents
        # threads from thinking a connection is complete when it is actually
        # still in progress.
        self.connected = conn
        return

    def wait_for_connection(self, timeout=None):
        """Wait for a connection that started with connect() to finish."""
        if not self.connected:
            start_time = time.time()
            if not self._conn_started:
                self.connect()

            if not self.connected:
                if timeout is None:
                    timeout = self.connection_timeout
                while not self.connected and time.time()-start_time < timeout:
                    time.sleep(0.1)
        return self.connected

    def connect(self, timeout=None):
        """Check that a PV is connected, forcing a connection if needed."""
        if not self.connected:
            if timeout is None:
                timeout = self.connection_timeout
        self._conn_started = True
        return self.connected and self.ftype is not None

    def clear_auto_monitor(self):
        """Turn off auto-monitoring: must reconnect to re-enable monitoring."""
        self.auto_monitor = False

    def reconnect(self):
        """Try to reconnect PV."""
        self.auto_monitor = None
        self.connected = False
        self._conn_started = False
        self.force_connect()
        return self.wait_for_connection()

    def poll(self, evt=1.e-4, iot=1.0):
        """Poll for changes."""
        pass

    def get(self, count=None, as_string=False, as_numpy=True,
            timeout=None, with_ctrlvars=False, use_monitor=True):
        """returns current value of PV.  Use the options:
        count       explicitly limit count for array data
        as_string   flag(True/False) to get a string representation
                    of the value.
        as_numpy    flag(True/False) to use numpy array as the
                    return type for array data.
        timeout     maximum time to wait for value to be received.
                    (default = 0.5 + log10(count) seconds)
        use_monitor flag(True/False) to use value from latest
                    monitor callback (True, default) or to make an
                    explicit CA call for the value.

        >>> p.get('13BMD:m1.DIR')
        0
        >>> p.get('13BMD:m1.DIR', as_string=True)
        'Pos'
        """
        if not self.wait_for_connection(timeout=timeout):
            return None
        if with_ctrlvars and getattr(self, 'units', None) is None:
            self.get_ctrlvars()

        val = self._args['value']
        if as_string:
            return self._set_charval(val)
        if self.count <= 1 or val is None:
            return val

        # After this point:
        #   * self.count is > 1
        #   * val should be set and a sequence
        try:
            len(val)
        except TypeError:
            # Edge case where a scalar value leaks through ca.unpack()
            val = [val]

        if count is None:
            count = len(val)

        if (as_numpy and HAS_NUMPY and
                not isinstance(val, _np.ndarray)):
            val = _np.asarray(val)
        elif (not as_numpy and HAS_NUMPY and
                isinstance(val, _np.ndarray)):
            val = val.tolist()
        # allow asking for less data than actually exists in the cached value
        if count < len(val):
            val = val[:count]
        return val

    def put(self, value, wait=False, timeout=30.0,
            use_complete=False, callback=None, callback_data=None):
        """Set value for PV.

        optionally waiting until the processing is
        complete, and optionally specifying a callback function to be run
        when the processing is complete.
        """
        if not self.wait_for_connection():
            return None

        if value is None:
            return None
        elif (self.ftype in (dbr.ENUM, dbr.TIME_ENUM, dbr.CTRL_ENUM) and
              isinstance(value, str)):
            if self._args['enum_strs'] is None:
                self.get_ctrlvars()
            if value in self._args['enum_strs']:
                # tuple.index() not supported in python2.5
                # value = self._args['enum_strs'].index(value)
                for ival, val in enumerate(self._args['enum_strs']):
                    if val == value:
                        value = ival
                        break
        elif (not isinstance(value, _np.ndarray) and
              not isinstance(value, self._pytype)):
            try:
                value = self._pytype(value)
            except Exception:
                return None
        if callback is None:
            callback = self.__putCallbackStub
        self.__on_changes(value=value)
        kws = {'pvname': self.pvname}
        kws.update(callback_data if callback_data else {})
        callback(**kws)

    def __putCallbackStub(self, pvname=None, **kws):
        """Null put-calback, so that the put_complete attribute is valid."""

    def _set_charval(self, val, call_ca=True):
        """Set the character representation of the value.

        intended only for internal use"""
        if val is None:
            self._args['char_value'] = 'None'
            return 'None'
        ftype = self._args['ftype']
        ntype = dbr.native_type(ftype)
        if ntype == dbr.STRING:
            self._args['char_value'] = val
            return val
        # char waveform as string
        if ntype == dbr.CHAR and self.count < AUTOMONITOR_MAXLENGTH:
            if HAS_NUMPY and isinstance(val, _np.ndarray):
                # a numpy array
                val = val.tolist()

                if not isinstance(val, list):
                    # a scalar value from numpy, tolist() turns it into a
                    # native python integer
                    val = [val.tolist()]
            else:
                try:
                    # otherwise, try forcing it into a list. this will fail for
                    # scalar types
                    val = list(val)
                except TypeError:
                    # and when it fails, make it a list of one scalar value
                    val = [val]

            if 0 in val:
                firstnull = val.index(0)
            else:
                firstnull = len(val)
            try:
                cval = ''.join([chr(i) for i in val[:firstnull]]).rstrip()
            except ValueError:
                cval = ''
            self._args['char_value'] = cval
            return cval

        cval = repr(val)
        if self.count > 1:
            try:
                length = len(val)
            except TypeError:
                length = 1
            cval = '<array size=%d, type=%s>' % (length,
                                                 dbr.Name(ftype).lower())
        elif ntype in (dbr.FLOAT, dbr.DOUBLE):
            if call_ca and self._args['precision'] is None:
                self.get_ctrlvars()
            try:
                prec = self._args['precision']
                fmt = "%%.%if"
                if 4 < abs(int(log10(abs(val + 1.e-9)))):
                    fmt = "%%.%ig"
                cval = (fmt % prec) % val
            except (ValueError, TypeError, ArithmeticError):
                cval = str(val)
        elif ntype == dbr.ENUM:
            if call_ca and self._args['enum_strs'] in ([], None):
                self.get_ctrlvars()
            try:
                cval = self._args['enum_strs'][val]
            except (TypeError, KeyError, IndexError):
                cval = str(val)

        self._args['char_value'] = cval
        return cval

    def get_ctrlvars(self, timeout=5, warn=True):
        """Get control values for variable."""
        if not self.wait_for_connection():
            return None
        kwds = dict()
        for key in ('precision', 'units', 'severity', 'status',
                    'upper_disp_limit', 'lower_disp_limit',
                    'upper_alarm_limit', 'upper_warning_limit',
                    'lower_warning_limit', 'lower_alarm_limit',
                    'upper_ctrl_limit', 'lower_ctrl_limit'):
            if self._args[key] is not None:
                kwds[key] = self._args[key]

        self.force_read_access_rights()
        return kwds

    def get_timevars(self, timeout=5, warn=True):
        "get time values for variable"
        if not self.wait_for_connection():
            return None
        kwds = dict()
        for key in ('status', 'severity', 'timestamp'):
            if self._args[key] is not None:
                kwds[key] = self._args[key]
        return kwds

    def __on_changes(self, value=None, **kwd):
        """Run internal callback function.

        Do not overwrite!!
        To have user-defined code run when the PV value changes,
        use add_callback()
        """
        self._args.update(kwd)
        self._args['value'] = value
        self._args['timestamp'] = time.time()
        self._args['posixseconds'] = 0
        self._args['nanoseconds'] = 0
        self._set_charval(self._args['value'], call_ca=False)
        if self.verbose:
            now = fmt_time(self._args['timestamp'])
            write('%s: %s (%s)' % (self.pvname,
                                   self._args['char_value'], now))
        self.run_callbacks()

    def run_callbacks(self):
        """Run all user-defined callbacks with the current data.

        Normally, this is to be run automatically on event, but
        it is provided here as a separate function for testing
        purposes.
        """
        for index in sorted(list(self.callbacks.keys())):
            self.run_callback(index)

    def run_callback(self, index):
        """Run a specific user-defined callback.

        specified by index, with the current data.
        Note that callback functions are called with keyword/val
        arguments including:
             self._args  (all PV data available, keys = __fields)
             keyword args included in add_callback()
             keyword 'cb_info' = (index, self)
        where the 'cb_info' is provided as a hook so that a callback
        function  that fails may de-register itself (for example, if
        a GUI resource is no longer available).
        """
        try:
            fcn, kwargs = self.callbacks[index]
        except KeyError:
            return
        kwd = copy.copy(self._args)
        kwd.update(kwargs)
        kwd['cb_info'] = (index, self)
        if hasattr(fcn, '__call__'):
            fcn(**kwd)

    def add_callback(self, callback=None, index=None, run_now=False,
                     with_ctrlvars=True, **kw):
        """Add a callback to a PV.

        Optional keyword arguments set here will be preserved and
        passed on to the callback at runtime.

        Note that a PV may have multiple callbacks, so that each
        has a unique index (small integer) that is returned by
        add_callback.  This index is needed to remove a callback.
        """
        if hasattr(callback, '__call__'):
            if index is None:
                index = 1
                if self.callbacks:
                    index = 1 + max(self.callbacks.keys())
            self.callbacks[index] = (callback, kw)

        if with_ctrlvars and self.connected:
            self.get_ctrlvars()
        if run_now:
            self.get(as_string=True)
            if self.connected:
                self.run_callback(index)
        return index

    def remove_callback(self, index=None):
        """Remove a callback by index."""
        if index in self.callbacks:
            self.callbacks.pop(index)

    def clear_callbacks(self):
        """Clear all callbacks."""
        self.callbacks = {}

    def _getinfo(self):
        """Get information paragraph."""
        if not self.wait_for_connection():
            return None
        self.get_ctrlvars()
        out = []
        mod = 'native'
        xtype = self._args['typefull']
        if '_' in xtype:
            mod, xtype = xtype.split('_')

        fmt = '%i'
        if   xtype in ('float','double'):
            fmt = '%g'
        elif xtype in ('string','char'):
            fmt = '%s'

        self._set_charval(self._args['value'], call_ca=False)
        out.append("== %s  (%s_%s) ==" % (self.pvname, mod, xtype))
        if self.count == 1:
            val = self._args['value']
            out.append('   value      = %s' % fmt % val)
        else:
            ext = {True: '...', False: ''}[self.count > 10]
            elems = range(min(5, self.count))
            try:
                aval = [fmt % self._args['value'][i] for i in elems]
            except TypeError:
                aval = ('unknown',)
            out.append("   value      = array  [%s%s]" % (",".join(aval), ext))
        for nam in ('char_value', 'count', 'nelm', 'type', 'units',
                    'precision', 'host', 'access',
                    'status', 'severity', 'timestamp',
                    'posixseconds', 'nanoseconds',
                    'upper_ctrl_limit', 'lower_ctrl_limit',
                    'upper_disp_limit', 'lower_disp_limit',
                    'upper_alarm_limit', 'lower_alarm_limit',
                    'upper_warning_limit', 'lower_warning_limit'):
            if hasattr(self, nam):
                att = getattr(self, nam)
                if att is not None:
                    if nam == 'timestamp':
                        att = "%.3f (%s)" % (att, fmt_time(att))
                    elif nam == 'char_value':
                        att = "'%s'" % att
                    if len(nam) < 12:
                        out.append('   %.11s= %s' % (nam+' '*12, str(att)))
                    else:
                        out.append('   %.20s= %s' % (nam+' '*20, str(att)))
        if xtype == 'enum':  # list enum strings
            out.append('   enum strings: ')
            for index, nam in enumerate(self.enum_strs):
                out.append("       %i = %s " % (index, nam))

        if self.auto_monitor is not None:
            msg = 'PV is internally monitored'
            out.append('   %s, with %i user-defined callbacks:' % (
                msg, len(self.callbacks)))
            if self.callbacks:
                for nam in sorted(self.callbacks.keys()):
                    cback = self.callbacks[nam][0]
                    out.append('      %s in file %s' % (
                        cback.func_name, cback.func_code.co_filename))
        else:
            out.append('   PV is NOT internally monitored')
        out.append('=============================')
        return '\n'.join(out)

    def _getarg(self, arg):
        "wrapper for property retrieval"
        if self._args['value'] is None:
            self.get()
        if self._args[arg] is None:
            if arg in ('status', 'severity', 'timestamp',
                       'posixseconds', 'nanoseconds'):
                self.get_timevars(timeout=1, warn=False)
            else:
                self.get_ctrlvars(timeout=1, warn=False)
        return self._args.get(arg, None)

    def __getval__(self):
        """get value."""
        return self._getarg('value')

    def __setval__(self, val):
        """Put value."""
        return self.put(val)

    value = property(__getval__, __setval__, None, "value property")

    @property
    def char_value(self):
        """Return character string representation of value."""
        return self._getarg('char_value')

    @property
    def status(self):
        """Return pv status."""
        return self._getarg('status')

    @property
    def type(self):
        """"Return pv type."""
        return self._args['type']

    @property
    def typefull(self):
        """Retunr full pv type."""
        return self._args['typefull']

    @property
    def host(self):
        """Return pv host."""
        return self._getarg('host')

    @property
    def count(self):
        """Return number of elements.

        For array data and later EPICS versions,
        this is equivalent to the .NORD field.  See also 'nelm' property.
        """
        if self._args['count'] is not None:
            return self._args['count']
        else:
            return self._getarg('count')

    @property
    def nelm(self):
        """Return native count (number of elements).

        For array data this will return the full array size (ie, the
        .NELM field).  See also 'count' property
        """
        if self._getarg('count') == 1:
            return 1
        return self._getarg('nelm')

    @property
    def read_access(self):
        """Reurn read access."""
        return self._getarg('read_access')

    @property
    def write_access(self):
        """Return write access."""
        return self._getarg('write_access')

    @property
    def access(self):
        """Return read/write access as string."""
        return self._getarg('access')

    @property
    def severity(self):
        """Return pv severity."""
        return self._getarg('severity')

    @property
    def timestamp(self):
        """Return timestamp of last pv action."""
        return self._getarg('timestamp')

    @property
    def posixseconds(self):
        """Return integer seconds for timestamp of last pv action.

        ...Using POSIX time convention.
        """
        return self._getarg('posixseconds')

    @property
    def nanoseconds(self):
        """Return integer nanoseconds for timestamp of last pv action."""
        return self._getarg('nanoseconds')

    @property
    def precision(self):
        """return number of digits after decimal point."""
        return self._getarg('precision')

    @property
    def units(self):
        """Return engineering units for pv."""
        return self._getarg('units')

    @property
    def enum_strs(self):
        """Return list of enumeration strings."""
        return self._getarg('enum_strs')

    @property
    def upper_disp_limit(self):
        """Return limit."""
        return self._getarg('upper_disp_limit')

    @property
    def lower_disp_limit(self):
        """Return limit."""
        return self._getarg('lower_disp_limit')

    @property
    def upper_alarm_limit(self):
        """Return limit."""
        return self._getarg('upper_alarm_limit')

    @property
    def lower_alarm_limit(self):
        """Return limit."""
        return self._getarg('lower_alarm_limit')

    @property
    def lower_warning_limit(self):
        """Return limit."""
        return self._getarg('lower_warning_limit')

    @property
    def upper_warning_limit(self):
        """Return limit."""
        return self._getarg('upper_warning_limit')

    @property
    def upper_ctrl_limit(self):
        """Return limit."""
        return self._getarg('upper_ctrl_limit')

    @property
    def lower_ctrl_limit(self):
        """Return limit."""
        return self._getarg('lower_ctrl_limit')

    @property
    def info(self):
        """Return info string."""
        return self._getinfo()

    @property
    def put_complete(self):
        """Return True if a put-with-wait has completed."""
        return True

    def __repr__(self):
        """Implement string representation."""
        if self.connected:
            if self.count == 1:
                return self._fmtsca % self._args
            else:
                return self._fmtarr % self._args
        else:
            return "<PV '%s': not connected>" % self.pvname

    def __eq__(self, other):
        """Implement test for equality."""
        try:
            return (self.chid == other.chid)
        except AttributeError:
            return False

    def disconnect(self):
        """Disconnect PV."""
        self.connected = False
        pvid = (self.pvname, self.form, self.context)
        if pvid in _PVCACHE_:
            _PVCACHE_.pop(pvid)

        if self.auto_monitor:
            self._args = {}.fromkeys(self._fields)

        self.callbacks = {}



"""Simulators factory."""

import re as _re

from ..namesys import split_name as _split_name

from .simps import SimPSTypeModel as _SimPSTypeModel
from .simps import SimPUTypeModel as _SimPUTypeModel


class _SimFactoryData:
    """."""

    # NOTE: To add a new simulator append an item in this dictionary
    # and implement corresponding method. Note that order in dictionary
    # is important!

    SIMCLASS_2_CREATE = {
        _SimPSTypeModel: '_simpstypemodel',
        _SimPUTypeModel: '_simpstypemodel',
        }

    @staticmethod
    def _simpstypemodel(simclass, pvname,):
        """Create simulator for Sim(PS|PU)TypeModel and return check."""
        # Create branch
        kwargs = _split_name(pvname)
        devname = kwargs['device_name']
        typemodel = simclass.conv_psname2typemodel(devname)
        simulator = simclass(*typemodel)
        pv_in_sim = simulator.pv_check(pvname)
        return simulator, pv_in_sim


class SimFactory(_SimFactoryData):
    """Simulations factory."""

    @staticmethod
    def create(pvnames, sims=None):
        """Create or add simulators to the oringinal set.

        Parameters
        pvnames : iterable with devnames and/or pvnames.
        sims : set of existing simulators to be updated.

        """
        if isinstance(pvnames, str):
            pvnames = [pvnames]

        if sims is None:
            sims = set()

        for pvname in pvnames:
            if not SimFactory._add_pvname(pvname, sims):
                raise ValueError('Simulator not found for "{}"'.format(pvname))

        return sims

    @staticmethod
    def _add_pvname(pvname, sims):
        simtypes = {type(sim) for sim in sims}
        for simclass, func in _SimFactoryData.SIMCLASS_2_CREATE.items():
            method = getattr(_SimFactoryData, func)
            # create simulator
            sim, pv_in_sim = method(simclass, pvname)
            if pv_in_sim:
                if simclass in simtypes:
                    # simulator type already in set
                    return True
                # add simulator to set
                sims.add(sim)
                simtypes.add(simclass)
                return True
        return False



"""Definition module."""
import math as _math
import logging as _log
import numpy as _np

from ..callbacks import Callback as _Callback

from .csdev import SOFBFactory as _SOFBFactory, ConstTLines as _ConstTLines

_CSACCELS = {}


def _create_csorb(acc):
    csorb = _CSACCELS.get(acc)
    if csorb is None:
        csorb = _SOFBFactory.create(acc)
        _CSACCELS[acc] = csorb
    return csorb


class BaseClass(_Callback):
    """Base Class."""

    def __init__(self, acc, prefix='', callback=None):
        """Init method."""
        super().__init__(callback)
        self._csorb = _create_csorb(acc)
        self._prefix = prefix
        self._status = 0b0
        self._map2write = self.get_map2write()

    @property
    def prefix(self):
        """Prefix."""
        return self._prefix

    @property
    def acc(self):
        """Accelerator name."""
        return self._csorb.acc

    @property
    def acc_idx(self):
        """Accelerator index."""
        return self._csorb.acc_idx

    @property
    def isring(self):
        """Ring accelerator status."""
        return self._csorb.isring

    @property
    def status(self):
        """Status."""
        self._update_status()
        return self._status

    @property
    def csorb(self):
        """CSDevice SOFB definition."""
        return self._csorb

    def write(self, pvname, value):
        """."""
        pvname = pvname.replace(self.prefix, '')
        if isinstance(value, (_np.ndarray, list, tuple)):
            pval = f'{value[0]}...{value[-1]}'
        else:
            pval = f'{value}'
        _log.info(f'Write received for: {pvname} --> {pval}')
        if pvname in self._map2write:
            ret = self._map2write[pvname](value)
            if ret:
                _log.info(f'YES Write for: {pvname} --> {pval}')
            else:
                _log.info(f'NOT Write for: {pvname} --> {pval}')
            return ret
        else:
            _log.warning('PV %s does not have a set function.', pvname)
            return False

    def run_callbacks(self, pvname, *args, **kwargs):
        """Run callback functions."""
        super().run_callbacks(self._prefix + pvname, *args, **kwargs)

    def get_map2write(self):
        """Return map of PV name to function for write."""
        return dict()

    def _update_log(self, value):
        self.run_callbacks('Log-Mon', value)

    def _update_status(self):
        pass


class BaseTimingConfig(_Callback):
    """Base timing configuration class."""

    def __init__(self, acc, callback=None):
        """Init method."""
        super().__init__(callback)
        self._csorb = _create_csorb(acc)
        self._config_ok_vals = {}
        self._config_pvs_rb = {}
        self._config_pvs_sp = {}
        self.put_enable = True

    @property
    def connected(self):
        """Status connected."""
        conn = True
        for pv in self._config_pvs_rb.values():
            if not pv.connected:
                _log.debug('NOT CONN: ' + pv.pvname)
            conn &= pv.connected

        for pv in self._config_pvs_sp.values():
            if not pv.connected:
                _log.debug('NOT CONN: ' + pv.pvname)
            conn &= pv.connected
        return conn

    @property
    def is_ok(self):
        """Ok status."""
        for k, val in self._config_ok_vals.items():
            pv = self._config_pvs_rb[k]
            pvval = None
            if pv.connected:
                pvval = pv.value
            if pvval is None:
                okay = False
                pvval = 'None'
            elif isinstance(val, float):
                okay = _math.isclose(val, pvval, rel_tol=1e-2)
            else:
                okay = val == pvval
            if not okay:
                msg = 'ERR: NOT CONF: {0:s}'.format(pv.pvname)
                self.run_callbacks('Log-Mon', msg)
                msg = msg[5:] + ' okv = {0}, v = {1}'.format(val, pvval)
                _log.warning(msg)
                return False
        return True

    def configure(self):
        """Configure method."""
        if not self.connected:
            return False

        if not self.put_enable:
            return True
        for k, pvo in self._config_pvs_sp.items():
            if k in self._config_ok_vals:
                pvo.put(self._config_ok_vals[k], wait=False)
        return True


def compare_kicks(val1, val2, atol=_ConstTLines.TINY_KICK, rtol=0):
    """."""
    if isinstance(val1, _np.ndarray) or isinstance(val2, _np.ndarray):
        return _np.isclose(val1, val2, atol=atol, rtol=rtol)
    return _math.isclose(val1, val2, abs_tol=atol, rel_tol=rtol)



"""Unittest module for ps_search.py."""

from unittest import TestCase

from siriuspy import util
from siriuspy.search import ps_search
from siriuspy.search import PSSearch
from siriuspy.pwrsupply.siggen import Signal
from siriuspy.magnet.excdata import ExcitationData


class TestModule(TestCase):
    """Test Search module."""

    public_interface = (
        'PSSearch',
    )

    def _test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            ps_search, TestModule.public_interface)
        self.assertTrue(valid)


class TestPSSearch(TestCase):
    """Test PSSearch."""

    public_interface = (
        'get_psnames',
        'get_psnicknames',
        'get_pstype_names',
        'get_psmodel_names',
        'get_bbbnames',
        'get_splims',
        'get_pstype_dict',
        'get_bbbname_dict',
        'get_udc_dict',
        'get_polarities',
        'conv_psname_2_pstype',
        'conv_psname_2_splims',
        'conv_pstype_2_polarity',
        'conv_pstype_2_magfunc',
        'conv_pstype_2_splims',
        'conv_psname_2_magfunc',
        'conv_psname_2_excdata',
        'conv_psname_2_psmodel',
        'conv_psmodel_2_psname',
        'conv_psname_2_siggenconf',
        'conv_psname_2_bbbname',
        'conv_bbbname_2_psnames',
        'conv_bbbname_2_bsmps',
        'conv_bbbname_2_freq',
        'conv_bbbname_2_udc',
        'conv_udc_2_bbbname',
        'conv_udc_2_bsmps',
        'conv_psname_2_udc',
        'conv_psname_2_dclink',
        'conv_dclink_2_psname',
        'get_linac_ps_sinap2sirius_dict',
        'get_linac_ps_sirius2sinap_dict',
        'get_pstype_2_psnames_dict',
        'get_pstype_2_splims_dict',
        'get_splims_unit',
        'get_splims_labels',
        'get_psname_2_dclink_dict',
        'get_dclink_2_psname_dict',
    )

    sample = {
        'SI-Fam:PS-B1B2-1': 'si-dipole-b1b2-fam',
        'SI-Fam:PS-QDA': 'si-quadrupole-q14-fam',
        'SI-Fam:PS-SDB2': 'si-sextupole-s15-sd-fam',
        'SI-01C1:PS-CH': 'si-sextupole-s15-ch',
        'SI-02C3:PS-CV-2': 'si-sextupole-s15-cv',
        'SI-03M1:PS-QS': 'si-sextupole-s15-qs',
        'SI-02M1:PS-FCV': 'si-corrector-fcv',
        'BO-48D:PU-EjeKckr': 'bo-ejekicker',
        'SI-01SA:PU-PingH': 'si-hping',
        'SI-01SA:PU-InjDpKckr': 'si-injdpk',
        'SI-01SA:PU-InjNLKckr': 'si-injnlk',
        'SI-19C4:PU-PingV': 'si-vping',
        'TB-04:PU-InjSept': 'tb-injseptum',
        'TS-01:PU-EjeSeptF': 'ts-ejeseptum-thin',
        'TS-01:PU-EjeSeptG': 'ts-ejeseptum-thick',
        'TS-04:PU-InjSeptF': 'ts-injseptum-thin',
        'TS-04:PU-InjSeptG-1': 'ts-injseptum-thick',
        'TS-04:PU-InjSeptG-2': 'ts-injseptum-thick',
    }

    sample_bbb = {
        'PA-RaPSC03:CO-PSCtrl-BO4': None,
        'PA-RaPSC03:CO-PSCtrl-BO3': None,
        'PA-RaPSC03:CO-PSCtrl-BO2': None,
        'PA-RaPSE05:CO-PSCtrl-BO': None,
    }

    pstype2polarity = {
        'si-dipole-b1b2-fam': 'monopolar',
        'si-quadrupole-q14-fam': 'monopolar',
        'si-sextupole-s15-sd-fam': 'monopolar',
        'si-sextupole-s15-ch': 'bipolar',
        'si-sextupole-s15-cv': 'bipolar',
        'si-sextupole-s15-qs': 'bipolar',
        'si-corrector-fcv': 'bipolar',
        'bo-ejekicker': 'monopolar',
        'si-hping': 'bipolar',
        'si-injdpk': 'monopolar',
        'si-injnlk': 'monopolar',
        'si-vping': 'bipolar',
        'tb-injseptum': 'monopolar',
        'ts-ejeseptum-thin': 'monopolar',
        'ts-ejeseptum-thick': 'monopolar',
        'ts-injseptum-thin': 'monopolar',
        'ts-injseptum-thick': 'monopolar',
    }

    pstype2magfunc = {
        'si-dipole-b1b2-fam': 'dipole',
        'si-quadrupole-q14-fam': 'quadrupole',
        'si-sextupole-s15-sd-fam': 'sextupole',
        'si-sextupole-s15-ch': 'corrector-horizontal',
        'si-sextupole-s15-cv': 'corrector-vertical',
        'si-sextupole-s15-qs': 'quadrupole-skew',
        'si-corrector-fcv': 'corrector-vertical',
        'bo-ejekicker': 'corrector-horizontal',
        'si-hping': 'corrector-horizontal',
        'si-injdpk': 'corrector-horizontal',
        'si-injnlk': 'corrector-horizontal',
        'si-vping': 'corrector-vertical',
        'tb-injseptum': 'corrector-horizontal',
        'ts-ejeseptum-thin': 'corrector-horizontal',
        'ts-ejeseptum-thick': 'corrector-horizontal',
        'ts-injseptum-thin': 'corrector-horizontal',
        'ts-injseptum-thick': 'corrector-horizontal',
    }

    def test_public_interface(self):
        """Test class public interface."""
        valid = util.check_public_interface_namespace(
            PSSearch, TestPSSearch.public_interface)
        self.assertTrue(valid)

    def test_get_psnames(self):
        """Test get_psnames."""
        # without filters
        psnames = PSSearch.get_psnames()
        self.assertIsInstance(psnames, (list, tuple))
        for psname in TestPSSearch.sample:
            self.assertIn(psname, psnames)
        # check sorted
        sorted_psnames = sorted(psnames)
        self.assertEqual(psnames, sorted_psnames)
        # with filters
        psnames = PSSearch.get_psnames({'dis': 'PU'})
        self.assertEqual(len(psnames), 12)
        for name in psnames:
            self.assertIn('PU', name)
        psnames = PSSearch.get_psnames({'sub': '0.M1'})
        self.assertEqual(len(psnames), 69)
        # exceptions
        self.assertRaises(TypeError, PSSearch.get_psnames, filters=23)
        self.assertRaises(TypeError, PSSearch.get_psnames, filters=23.4)
        self.assertRaises(TypeError, PSSearch.get_psnames, filters=[0, ])
        self.assertRaises(TypeError, PSSearch.get_psnames, filters=(0.0, ))

    def test_get_pstype_names(self):
        """Test get_pstype_names."""
        pstypes = PSSearch.get_pstype_names()
        self.assertIsInstance(pstypes, list)
        for pstype in pstypes:
            self.assertIsInstance(pstype, str)

    def test_get_bbbnames(self):
        """Test get_bbbnames."""
        # without filters
        bbbnames = PSSearch.get_bbbnames()
        self.assertIsInstance(bbbnames, (list, tuple))
        for bbbname in TestPSSearch.sample_bbb:
            self.assertIn(bbbname, bbbnames)
        # check sorted
        sorted_bbbnames = sorted(bbbnames)
        self.assertEqual(bbbnames, sorted_bbbnames)
        # # with filters
        # bbbnames = PSSearch.get_bbbnames({'dis': 'CO'})
        # self.assertEqual(len(bbbnames), 289)
        # for name in bbbnames:
        #     self.assertIn('CO', name)
        # bbbnames = PSSearch.get_bbbnames({'sub': 'Glob'})
        # self.assertEqual(len(bbbnames), 29)
        # exceptions
        self.assertRaises(TypeError, PSSearch.get_bbbnames, filters=23)
        self.assertRaises(TypeError, PSSearch.get_bbbnames, filters=23.4)
        self.assertRaises(TypeError, PSSearch.get_bbbnames, filters=[0, ])
        self.assertRaises(TypeError, PSSearch.get_bbbnames, filters=(0.0, ))

    def test_get_splims(self):
        """Test get_splims."""
        lim1 = PSSearch.get_splims(
            pstype='si-quadrupole-q30-trim', label='lolo')
        lim2 = PSSearch.get_splims(
            pstype='si-quadrupole-q30-trim', label='hihi')
        self.assertGreater(lim2, lim1)
        # exceptions
        self.assertRaises(
            KeyError, PSSearch.get_splims,
            pstype='dummy', label='low')
        self.assertRaises(
            KeyError, PSSearch.get_splims,
            pstype='bo-corrector-ch', label='dummy')

    def test_get_pstype_dict(self):
        """Test get_pstype_dict."""
        dict_ = PSSearch.get_pstype_dict()
        self.assertIsInstance(dict_, dict)
        pstypes_d = sorted(list(dict_.keys()))
        pstypes = sorted(PSSearch.get_pstype_names())
        self.assertEqual(pstypes_d, pstypes)

    def test_get_bbbname_dict(self):
        """Test get_bbbname_dict."""
        dict_ = PSSearch.get_bbbname_dict()
        self.assertIsInstance(dict_, dict)
        for bbbname in TestPSSearch.sample_bbb:
            self.assertTrue(bbbname in dict_)

    def test_get_polarities(self):
        """Test get_polarities."""
        polarities = PSSearch.get_polarities()
        self.assertIsInstance(polarities, list)
        for pol in polarities:
            self.assertIsInstance(pol, str)
        self.assertIn('bipolar', polarities)
        self.assertIn('monopolar', polarities)

    def test_conv_psname_2_pstypes(self):
        """Test conv_psname_2_pstype."""
        for psname, pstype in TestPSSearch.sample.items():
            self.assertEqual(PSSearch.conv_psname_2_pstype(psname), pstype)
        # exceptions
        self.assertRaises(
            KeyError, PSSearch.conv_psname_2_pstype, psname='dummy')

    def test_conv_pstype_2_polarity(self):
        """Test conv_pstype_2_polarity."""
        for pstype, polarity in TestPSSearch.pstype2polarity.items():
            self.assertEqual(
                PSSearch.conv_pstype_2_polarity(pstype), polarity)
        # Non existent pstype raises KeyError
        self.assertRaises(KeyError,
                          PSSearch.conv_pstype_2_polarity, pstype='dummy')

    def test_conv_pstype_2_magfunc(self):
        """Test conv_pstype_2_polarity."""
        for pstype, magfunc in TestPSSearch.pstype2magfunc.items():
            self.assertEqual(
                PSSearch.conv_pstype_2_magfunc(pstype), magfunc)
        # Non existent pstype raises KeyError
        self.assertRaises(KeyError,
                          PSSearch.conv_pstype_2_magfunc, pstype='dummy')

    def test_conv_pstype_2_splims(self):
        """Test conv_pstype_2_polarity."""
        pstypes = tuple(TestPSSearch.pstype2polarity.keys())
        for pstype in pstypes:
            splims = PSSearch.conv_pstype_2_splims(pstype)
            self.assertIsInstance(splims, dict)
        # Non existent pstype return None
        self.assertRaises(
            KeyError, PSSearch.conv_pstype_2_splims, pstype='dummy')

    def test_conv_psname_2_excdata(self):
        """Test conv_psname_2_excdata."""
        self.assertRaises(KeyError, PSSearch.conv_psname_2_excdata, '')
        excdata = PSSearch.conv_psname_2_excdata('SI-Fam:PS-B1B2-1')
        self.assertIsInstance(excdata, ExcitationData)

    def test_conv_psname_2_psmodel(self):
        """Test conv_psname_2_psmodel."""
        for psn in TestPSSearch.sample:
            model = PSSearch.conv_psname_2_psmodel(psname=psn)
            self.assertIsInstance(model, str)

    def test_conv_psname_2_siggenconf(self):
        """Test conv_psname_2_siggenconf."""
        for psn in TestPSSearch.sample:
            siggenconf = PSSearch.conv_psname_2_siggenconf(psname=psn)
            self.assertIsInstance(siggenconf, Signal)

    def test_conv_bbbname_2_psnames(self):
        """Test conv_bbbname_2_psnames."""
        self.assertRaises(TypeError, PSSearch.conv_bbbname_2_psnames)
        self.assertRaises(KeyError, PSSearch.conv_bbbname_2_psnames, '')
        bsmps = PSSearch.conv_bbbname_2_psnames(
            bbbname='PA-RaPSF05:CO-PSCtrl-BO')
        self.assertIsInstance(bsmps, list)
        self.assertGreater(len(bsmps), 0)
        self.assertIsInstance(bsmps[0], tuple)
        self.assertIsInstance(bsmps[0][0], str)
        self.assertIsInstance(bsmps[0][1], int)

    def test_conv_bbbname_2_freq(self):
        """Test conv_bbbname_2_freqs."""
        self.assertRaises(TypeError, PSSearch.conv_bbbname_2_freq)
        self.assertRaises(KeyError, PSSearch.conv_bbbname_2_psnames, '')
        freq = PSSearch.conv_bbbname_2_freq(
            bbbname='LA-RaCtrl:CO-PSCtrl-TB1')
        self.assertIsInstance(freq, float)

    def test_get_pstype_2_psnames_dict(self):
        """Test get_pstype_2_psnames_dict."""
        typ2name = PSSearch.get_pstype_2_psnames_dict()
        self.assertIsInstance(typ2name, dict)
        for pstype, psnames in typ2name.items():
            self.assertIsInstance(pstype, str)
            self.assertIsInstance(psnames, (tuple, list))
            self.assertTrue(len(psnames) > 0)

    def test_get_psname_2_splims_dict(self):
        """Test conv psname_2_splims_dict."""
        limlabels = ('DRVL', 'LOLO', 'LOW', 'LOPR',
                     'HOPR', 'HIGH', 'HIHI', 'DRVH')
        splims_dict = PSSearch.get_pstype_2_splims_dict()
        self.assertIsInstance(splims_dict, dict)
        for pstype, splims in splims_dict.items():
            self.assertIsInstance(pstype, str)
            self.assertIsInstance(splims, dict)
            for limlabel in limlabels:
                self.assertIn(limlabel, splims)
            self.assertTrue(splims['LOLO'] <= splims['LOW'])
            self.assertTrue(splims['LOW'] < splims['HIGH'])
            self.assertTrue(splims['HIGH'] <= splims['HIHI'])

    def test_get_splims_unit(self):
        """Test get_splims_unit."""
        self.assertRaises(TypeError, PSSearch.get_splims_unit)
        self.assertRaises(ValueError, PSSearch.get_splims_unit, '')
        self.assertRaises(TypeError, PSSearch.get_splims_unit)
        self.assertEqual(PSSearch.get_splims_unit('FBP'), ['A', 'Ampere'])

    def test_get_splims_labels(self):
        """Test get_splims_labels."""
        self.assertEqual(PSSearch.get_splims_labels(),
                         ['DRVL', 'LOLO', 'LOW', 'LOPR',
                          'HOPR', 'HIGH', 'HIHI', 'DRVH',
                          'DTOL_CUR', 'DTOL_WFM', 'TSTV', 'TSTR'])



"""Controller classes."""

from ...search import PSSearch as _PSSearch


class PSController:
    """Power supply controller.

    Objects of this class are used to communicate with power supply devices
    by invoking properties and methods of the PRU controller.
    """

    _FIELDS_IGNORED = {
        # these are fields managed in BeagleBone class objects.
        'Energy-SP', 'Energy-RB', 'EnergyRef-Mon', 'Energy-Mon',
        'Kick-SP', 'Kick-RB', 'KickRef-Mon', 'Kick-Mon',
        'KL-SP', 'KL-RB', 'KLRef-Mon', 'KL-Mon',
        'SL-SP', 'SL-RB', 'SLRef-Mon', 'SL-Mon'}

    def __init__(self, readers, writers,
                 pru_controller, devname2devid):
        """Initialize PS controller.

        readers : objects from classes in psreaders module that are responsible
                  for reading power supply parameters
        writers : objects from classes in pscwriters

        """
        self._readers = readers
        self._writers = writers
        self._pru_controller = pru_controller

        self._devname2devid = devname2devid

        # get all controller fields
        self._fields = self._get_fields()

        # shortcut to fields not to be ignored
        self._fields_not_ignored = self._fields - self._FIELDS_IGNORED

        # get udc to dev dictionary
        self._udc2dev = self._get_udc2dev()

    @property
    def pru_controller(self):
        """PRU controller."""
        return self._pru_controller

    @property
    def fields(self):
        """Field of ps controller."""
        return self._fields

    def read(self, devname, field):
        """Read pv value."""
        pvname = devname + ':' + field
        if field == 'CtrlLoop-Sts':
            sts = self._readers[devname + ':CtrlLoop-Sts']
            sel = self._readers[devname + ':CtrlLoop-Sel']
            if sts.read() != sel.read():
                sel.apply(sts.read())

        reader = self._readers[pvname]
        if reader is not None:
            return reader.read()
        raise AttributeError('Could not find reader for "{}"'.format(field))

    def read_all_fields(self, devname):
        """Read non-ignored-field values from device."""
        values = dict()
        for field in self._fields_not_ignored:
            pvname = devname + ':' + field
            value = self.read(devname, field)
            values[pvname] = value
        return values

    def write(self, devname, field, value):
        """Write value to pv."""
        priority_pvs = dict()

        pvname = devname + ':' + field
        if pvname in self._writers:
            self._writers[pvname].execute(value)

        # update all setpoint properties upon return from SOFBMode
        if value == 0 and 'SOFBMode-Sel' in field:
            self._update_setpoints(devname)

        # return priority pvs
        return priority_pvs

    def check_connected(self, devname):
        """Check if device is connected."""
        devid = self._devname2devid[devname]
        return self._pru_controller.check_connected(devid)

    def init_setpoints(self, devname=None):
        """Initialize controller setpoint fields."""
        for key, reader_sp in self._readers.items():

            # if devname was passed, continue in case reader does not belong to device
            if devname and devname not in key:
                continue

            # ignore non-setpoint fields
            if not key.endswith(('-Sel', '-SP')):
                continue

            # ignore strength fields
            *_, prop = key.split(':')
            if prop in PSController._FIELDS_IGNORED:
                continue

            # initiliaze setpoint fields using corresponding readers

            # use corresponding readback field
            field_rb = PSController._get_readback_field(key)

            # check if reader exists
            reader_rb = self._readers[field_rb]
            if reader_rb is None:
                raise AttributeError(
                    'Could not find reader for "{}"'.format(field_rb))

            # read readback value
            value = reader_rb.read()
            if key.endswith('OpMode-Sel'):
                # OpModel-Sel is shifted in 3 units relative OpMode-Sts
                if value is not None:
                    value = 0 if value < 3 else value - 3

            # apply value to setpoint using its reader
            reader_sp.apply(value)

    # --- private methods ---

    def _get_udc2dev(self):
        udc2dev = dict()
        for devname in self._devname2devid:
            udc = _PSSearch.conv_psname_2_udc(devname)
            if udc not in udc2dev:
                udc2dev[udc] = set()
            udc2dev[udc].add(devname)
        return udc2dev

    def _get_fields(self):
        fields = set()
        for name in self._readers:
            split = name.split(':')
            fields.add(split[-1])
        return fields

    def _update_setpoints(self, devname):
        """."""
        self.read_all_fields(devname)
        self.init_setpoints(devname)
        
    @staticmethod
    def _get_readback_field(field):
        # NOTE: to be updated
        return field.replace('-Sel', '-Sts').replace('-SP', '-RB')


class StandardPSController(PSController):
    """Standard PSController.

    This is used in DCDC-type power supply models.
    """

    _SIGGEN_PARMS = [
        'CycleType-Sel',
        'CycleNrCycles-SP',
        'CycleFreq-SP',
        'CycleAmpl-SP',
        'CycleOffset-SP',
        'CycleAuxParam-SP',  # start index of auxparams
    ]

    def write(self, devname, field, value):
        """Write value to pv."""
        priority_pvs = dict()

        pvname = devname + ':' + field
        if pvname not in self._writers:
            return priority_pvs

        if field == 'SOFBCurrent-SP':
            self._set_sofb_current(pvname, value, devname, field, priority_pvs)
        elif field in StandardPSController._SIGGEN_PARMS:
            self._set_siggen(pvname, value, devname, field, priority_pvs)
        else:
            self._writers[pvname].execute(value)

        # update all setpoint properties upon return from SOFBMode
        if 'SOFBMode-Sel' in field and value == 0:
            self._update_setpoints(devname)

        # return priority pvs
        return priority_pvs

    # --- private methods ---

    def _set_siggen(self, pvname, value, devname, field, priority_pvs):
        _ = priority_pvs
        idx = StandardPSController._SIGGEN_PARMS.index(field)
        values = self._get_siggen_arg_values(devname)
        if field == 'CycleAuxParam-SP':
            values[idx:] = value
        else:
            values[idx] = value
        self._writers[pvname].execute(values)

    def _set_sofb_current(self, pvname, value, devname, field, priority_pvs):
        _ = field

        # set actual SOFBCurrent-SP
        self._writers[pvname].execute(value)

        # add readback SOFBCurrent PVs (same device)
        for suffix in ('-RB', 'Ref-Mon', '-Mon'):
            pvn = pvname.replace('-SP', suffix)
            reader = self._readers[pvn]
            priority_pvs[pvn] = reader.read()

        # add priority SOFBCurrent-SP for other
        # devices in the same UDC.
        for udc_devnames in self._udc2dev.values():
            if devname in udc_devnames:
                # loop over other UDC devices
                for devname_ in udc_devnames:
                    if devname_ != devname:
                        pvn = devname_ + ':SOFBCurrent-SP'
                        priority_pvs[pvn] = value

    def _get_siggen_arg_values(self, devname):
        """Get cfg_siggen args."""
        args = [self._readers[devname + ':' + arg].read()
                for arg in StandardPSController._SIGGEN_PARMS[:-1]]
        aux = StandardPSController._SIGGEN_PARMS[-1]
        args.extend(self._readers[devname + ':' + aux].read())
        return args



"""BSMP exceptions."""


class SerialError(Exception):
    """Exception raised when there's a problem with serial."""


class SerialErrEmpty(SerialError):
    """Exception raised when empty message."""


class SerialErrCheckSum(SerialError):
    """Exception raised when there'a problem with checksum."""


class SerialErrPckgLen(SerialError):
    """Exception raised when there'a problem with package length."""


class SerialErrMsgShort(SerialError):
    """Exception raised when message is too short."""


class SerialAnomResp(SerialError):
    """Exception raised when response is not the expected one."""



"""Implementation of namesys functions nad classes."""

import types as _types
import re as _re


_ATTRS = (
    'channel_type',
    'prefix',
    'sec',
    'sub',
    'dis',
    'dev',
    'idx',
    'propty',
    'field',
    'device_name',
    'area_name',
    'propty_name',
    'propty_suffix',
    'device_propty',
)


def get_siriuspvname_attrs():
    """Return SiriusPVName attributes."""
    return [attr for attr in _ATTRS]


def join_name(**kwargs):
    """Return SiriusPVName object.

    Parameters
        channel_type [str] : PyDM channel type, ex: 'ca'
        prefix [str] : Prefix, ex: 'fac-454lnls'
        sec [str] : Section, ex: 'SI'
        sub [str] : Subsection, ex: 'Glob'
        dis [str] : Discipline, ex: 'PS'
        dev [str] : Device, ex: 'B'
        idx [str] : Device index, ex: '1'
        propty_name [str] : Property name, ex: 'Properties'
        propty_suffix [str] : Property suffix, ex: 'Cte'
        propty [str] : Full property, ex: 'Properties-Cte'
        field [str] : Field, ex: 'STATUS'

    """
    dic = {k: v for k, v in kwargs.items() if v}  # get valid args
    name = ''
    if len(dic.keys()) == 1:
        if 'propty' in dic.keys():
            name = dic['propty']
        elif 'propty_name' in dic.keys():
            name = dic['propty_name']
    elif len(dic.keys()) == 2:
        if 'sec' in dic.keys() and 'sub' in dic.keys():
            name = dic['sec'].upper() + '-' + dic['sub']
        elif 'dis' in dic.keys() and 'dev' in dic.keys():
            name = dic['dis'].upper() + '-' + dic['dev']
        elif 'propty_name' in dic.keys() and 'propty_suffix' in dic.keys():
            name = dic['propty_name'].upper() + '-' + dic['propty_suffix']
    elif len(dic.keys()) == 3:
        if 'dis' in dic.keys() and 'dev' in dic.keys() and 'idx' in dic.keys():
            name = dic['dis'].upper() + '-' + dic['dev'] + '-' + dic['idx']
    elif len(dic.keys()) > 3:
        name = dic['channel_type'] + '://' if 'channel_type' in dic.keys() else ''
        name = dic['prefix'] + '-' if 'prefix' in dic.keys() else ''
        name += (dic['sec'].upper() + '-' + dic['sub'] + ':' +
                 dic['dis'].upper() + '-' + dic['dev'])
        name += ('-' + dic['idx']) if 'idx' in dic.keys() else ''
        if 'propty_name' in dic.keys() and 'propty_suffix' in dic.keys():
            name += ':' + dic['propty_name'] + '-' + dic['propty_suffix']
            name += ('.' + dic['field']) if 'field' in dic.keys() else ''
        elif 'propty' in dic.keys():
            name += ':' + dic['propty']
            name += ('.' + dic['field']) if 'field' in dic.keys() else ''
        elif 'propty_name' in dic.keys():
            name += ':' + dic['propty_name']

    if not name:
        raise TypeError('Not a valid SiriusPVName elements set!')
    return SiriusPVName(name)


def split_name(pvname, elements=None):
    """Return dict with PV name split into fields.

    Parameters
        pvname [str] : a complete pvname or a valid part of it.
        elements [None, 'propty', 'sec-sub', 'dis-dev'] : if pvname is not a
            complete name, 'elements' says which part of pvname
            it corresponds to.

    """
    if not elements:
        elements = 'propty'

    # empty dictionary
    dic_ = {attr: '' for attr in _ATTRS}

    # strip PV name
    pvname = pvname.strip()
    if not pvname:
        return dic_

    # deals with channel
    dic_['channel_type'] = ''
    names = pvname.split('://')
    if len(names) == 2:
        dic_['channel_type'] = names[0]
        pvname = names[1]

    list_ = pvname.split(':')

    if len(list_) == 1:
        slist_ = list_[0].split('-')
        if len(slist_) == 1:
            dic_['propty'] = slist_[0]
            dic_['propty_name'] = slist_[0]
        elif len(slist_) == 2:
            if elements == 'propty':
                dic_['propty'] = slist_[0] + '-' + slist_[1]
                dic_['propty_name'] = slist_[0]
                dic_['propty_suffix'] = slist_[1]
            elif elements == 'sec-sub':
                dic_['sec'] = slist_[0]
                dic_['sub'] = slist_[1]
                dic_['area_name'] = slist_[0] + '-' + slist_[1]
            elif elements == 'dis-dev':
                dic_['dis'] = slist_[0]
                dic_['dev'] = slist_[1]
        elif len(slist_) == 3:
            dic_['dis'] = slist_[0]
            dic_['dev'] = slist_[1]
            dic_['idx'] = slist_[2]
    else:
        slist_ = list_[0].split('-')
        dic_['prefix'] = '-'.join([s for s in slist_[:-2]])
        dic_['area_name'] = '-'.join([s for s in slist_[-2:]])
        dic_['device_name'] = dic_['area_name'] + ':' + list_[1]

        dic_['sec'] = slist_[-2]
        dic_['sub'] = slist_[-1]

        slist_ = list_[1].split('-')
        dic_['dis'] = slist_[0]
        dic_['dev'] = slist_[1]
        dic_['idx'] = slist_[2] if len(slist_) >= 3 else ''

        if len(list_) > 2:
            slist_ = list_[2].split('.')
            sslist_ = slist_[0].split('-')
            dic_['propty'] = slist_[0]
            dic_['propty_name'] = sslist_[0]
            dic_['propty_suffix'] = sslist_[1] if len(sslist_) > 1 else ''
            dic_['field'] = slist_[1] if len(slist_) >= 2 else ''

        dic_['device_propty'] = (
            dic_['dev'] +
            ('-' + dic_['idx'] if dic_['idx'] else '') +
            (':' + dic_['propty'] if dic_['propty'] else '') +
            ('.' + dic_['field'] if dic_['field'] else ''))
    return dic_


def get_pair_sprb(pv_propty):
    """Return the equivalent [setpoint, readback] SiriusPVName property pair.

    Input: a SiriusPVName property, with a setpoint or a readback suffix.
    Output: the equivalent Sirius [setpoint, readback] pair.
    """
    _sp_rb = {'SP': 'RB', 'Sel': 'Sts'}
    _rb_sp = {v: k for k, v in _sp_rb.items()}

    suffix = pv_propty.propty_suffix

    if suffix in _sp_rb.keys():
        return [pv_propty, pv_propty.substitute(propty_suffix=_sp_rb[suffix])]
    elif suffix in _rb_sp.keys():
        return [pv_propty.substitute(propty_suffix=_rb_sp[suffix]), pv_propty]
    else:
        raise TypeError('Input is not a setpoint/readback property!')


class SiriusPVName(str):
    """Sirius PV Name Class."""

    def __new__(cls, pv_name, elements=None):
        """Implement new method."""
        name = split_name(pv_name, elements)
        obj = super().__new__(cls, pv_name)
        obj.channel_type = name['channel_type']
        obj.prefix = name['prefix']
        obj.sec = name['sec']
        obj.sub = name['sub']
        obj.area_name = name['area_name']
        obj.dis = name['dis']
        obj.dev = name['dev']
        obj.idx = name['idx']
        obj._device_name = name['device_name']
        obj.propty_name = name['propty_name']
        obj.propty_suffix = name['propty_suffix']
        obj.propty = name['propty']
        obj.device_propty = name['device_propty']
        obj.field = name['field']
        return obj

    @property
    def device_name(self):
        """."""
        return SiriusPVName(self._device_name)

    def substitute(self, **kwargs):
        """Return new SiriusPVName object with the atttributes changed."""
        dic_ = {}
        dic_['sec'] = self.sec
        dic_['sub'] = self.sub
        dic_['dis'] = self.dis
        dic_['dev'] = self.dev
        dic_['idx'] = self.idx
        dic_['propty'] = self.propty
        dic_['propty_name'] = self.propty_name
        dic_['propty_suffix'] = self.propty_suffix
        dic_['field'] = self.field
        dic_['prefix'] = self.prefix
        dic_['channel_type'] = self.channel_type
        dic_.update({k: v for k, v in kwargs.items() if isinstance(v, str)})
        if 'propty' in kwargs:
            dic_.pop('propty_name', None)
            dic_.pop('propty_suffix', None)
        return join_name(**dic_)

    def get_nickname(self, sec=False, dev=False):
        """."""
        nickname = ''
        if sec:
            nickname += self.sec + '-'
        nickname += self.sub
        if dev:
            nickname += ':' + self.dev
        nickname += '-' + self.idx if self.idx else ''
        return nickname

    def __lt__(self, other):
        """Less-than operator."""
        cond = ((type(other) == type(self)) and
                (self.sec == other.sec) and
                (self.sub != other.sub))
        return self._subsec_comp(other) if cond else super().__lt__(other)

    def __gt__(self, other):
        """Greater-than operator."""
        return other.__lt__(self)

    def __le__(self, other):
        """Less-or-equal operator."""
        return self.__lt__(other) or self.__eq__(other)

    def __ge__(self, other):
        """Greater-or-equal operator."""
        return self.__gt__(other) or self.__eq__(other)

    def strip(self, *args, **kwargs):
        """."""
        return SiriusPVName(super().strip(*args, **kwargs))

    def replace(self, *args, **kwargs):
        """."""
        return SiriusPVName(super().replace(*args, **kwargs))

    def _subsec_comp(self, other):
        """Subsection comparison."""
        my_ssec = self.sub
        th_ssec = other.sub
        my_dev = self.dev
        th_dev = other.dev
        if my_ssec == 'Glob':
            return False
        elif th_ssec == 'Glob':
            return True
        elif my_ssec == 'Fam':
            return False
        elif th_ssec == 'Fam':
            return True
        elif my_ssec == '01M1':
            if th_ssec == '01SA' and th_dev == 'ScrapH':
                return True
            else:
                return False
        elif th_ssec == '01M1':
            if my_ssec == '01SA' and my_dev == 'ScrapH':
                return False
            else:
                return True
        elif my_ssec == '01U' and my_dev in {'CV', 'BPM'}:
            return False
        elif th_ssec == '01U' and th_dev in {'CV', 'BPM'}:
            return True
        elif my_ssec[:2] != th_ssec[:2]:
            return my_ssec[:2] < th_ssec[:2]
        elif len(my_ssec) == 2:
            return False
        elif len(th_ssec) == 2:
            return True
        if my_ssec[2] == th_ssec[2]:
            return my_ssec[3] < th_ssec[3]
        elif my_ssec[2] == 'C':
            return False
        elif th_ssec[2] == 'C':
            return True
        elif my_ssec[2] == 'S':
            if th_ssec[2:4] == 'M1':
                return False
            elif th_ssec[2:4] == 'M2':
                return True
        elif th_ssec[2] == 'S':
            if my_ssec[2:4] == 'M1':
                return True
            elif my_ssec[2:4] == 'M2':
                return False

    @staticmethod
    def is_write_pv(pvname):
        """."""
        return SiriusPVName.is_sp_pv(pvname) or SiriusPVName.is_cmd_pv(pvname)

    @staticmethod
    def is_sp_pv(pvname):
        """."""
        return pvname.endswith(('-Sel', '-SP'))

    @staticmethod
    def is_cmd_pv(pvname):
        """."""
        return pvname.endswith('-Cmd')

    @staticmethod
    def is_cte_pv(pvname):
        """."""
        return pvname.endswith('-Cte')

    @staticmethod
    def is_rb_pv(pvname):
        """."""
        return pvname.endswith(('-Sts', '-RB'))

    @staticmethod
    def from_sp2rb(pvname):
        """."""
        return pvname.replace('-SP', '-RB').replace('-Sel', '-Sts')

    @staticmethod
    def from_rb2sp(pvname):
        """."""
        return pvname.replace('-RB', '-SP').replace('-Sts', '-Sel')


class Filter:
    """Filter class."""

    # NOTE: Are these class constants really useful?

    # PVName regex filters
    patterns = _types.SimpleNamespace()
    patterns.FAM = 'Fam'
    patterns.TRIM = '\d{2}\w{0,2}'
    patterns.DIPOLE = 'B.*'
    patterns.QUADRUPOLE = '(?:QD|QF|Q[0-9]).*'
    patterns.QUADRUPOLE_SKEW = 'QS'
    patterns.QD = 'QD.*'
    patterns.QF = 'QF.*'
    patterns.SEXTUPOLE = 'S(?:D|F)*'
    patterns.SD = 'SD.*'
    patterns.SF = 'SF.*'
    patterns.CORRECTOR = '(?:C|FC).*'
    patterns.SLOW_CHV = 'C(?:H|V).*'
    patterns.SLOW_CH = 'CH.*'
    patterns.SLOW_CV = 'CV.*'
    patterns.FAST_CHV = 'FC.*'
    patterns.FAST_CH = 'FCH.*'
    patterns.FAST_CV = 'FCV.*'

    filters = _types.SimpleNamespace()
    filters.FAM = {'sub': patterns.FAM}
    filters.TRIM = {'sub': patterns.TRIM}
    filters.DIPOLE = {'dev': patterns.DIPOLE}
    filters.QUADRUPOLE = {'dev': patterns.QUADRUPOLE}
    filters.QUADRUPOLE_SKEW = {'dev': patterns.QUADRUPOLE_SKEW}
    filters.QD = {'dev': patterns.QD}
    filters.QF = {'dev': patterns.QF}
    filters.SEXTUPOLE = {'dev': patterns.SEXTUPOLE}
    filters.SD = {'dev': patterns.SD}
    filters.SF = {'dev': patterns.SF}
    filters.CORRECTOR = {'dev': patterns.CORRECTOR}
    filters.SLOW_CHV = {'dev': patterns.SLOW_CHV}
    filters.SLOW_CH = {'dev': patterns.SLOW_CH}
    filters.SLOW_CV = {'dev': patterns.SLOW_CV}
    filters.FAST_CHV = {'dev': patterns.FAST_CHV}
    filters.FAST_CH = {'dev': patterns.FAST_CH}
    filters.FAST_CV = {'dev': patterns.FAST_CV}

    @staticmethod
    def add_filter(filters=None, sec=None, sub=None,
                   dis=None, dev=None):
        """."""
        if filters is None:
            filters = []
        fil = {}
        if sec is not None:
            fil['sec'] = sec
        if sub is not None:
            fil['sub'] = sub
        if dis is not None:
            fil['dis'] = dis
        if dev is not None:
            fil['dev'] = dev
        if fil:
            filters.append(fil)
        return filters

    @staticmethod
    def process_filters(pvnames, filters=None, sorting=None):
        """Return a sorted and filtered list of given pv name lists.

        'filters' is either a dictionary of a list of dictionaries whose keys
        are pv sub parts and the values are the desired patterns
        """
        if filters is None:
            return pvnames
        if isinstance(filters, dict):
            filters = [filters]
        elif isinstance(pvnames, str):
            pvnames = [pvnames]

        # build filter regexp
        fs = []
        for fil in filters:
            if 'sec' not in fil or fil['sec'] is None:
                fil['sec'] = '[A-Z]{2,4}'
            if 'sub' not in fil or fil['sub'] is None:
                fil['sub'] = '\w{2,16}'
            if 'dis' not in fil or fil['dis'] is None:
                fil['dis'] = '[A-Z]{2,6}'
            if 'dev' not in fil or fil['dev'] is None:
                fil['dev'] = '.+'
            pattern = (fil['sec'] + '-' + fil['sub'] + ':' +
                       fil['dis'] + '-' + fil['dev'])
            if 'idx' in fil:
                pattern += '-' + fil['idx']
            regexp = _re.compile(pattern)
            fs.append(regexp)

        # filter list
        filtered_list = list()
        for pvname in pvnames:
            for pattern in fs:
                if pattern.match(pvname):
                    filtered_list.append(pvname)
                    break

        if sorting is None:
            sorted_filtered_list = filtered_list
        elif sorting == 'length':
            raise NotImplementedError
        return sorted_filtered_list
